
import argparse
import sys
import traceback
import datetime
from typing import List

HELP_MSG = '''
surfing <command> [<args>]

The most commonly used commands are:
    init_raw                setup raw database
    init_basic              setup basic database
    init_derived            setup derived database
    init_view               setup view database
    download_raw            download raw data
    update_basic            update basic data base on raw data
    update_derived          update derived data based on basic and raw data
    update_view             update view data based on basic and derived data
    data_update             run download_raw, update_basic, update_derived and update_view in turn
    check_fund_list         check fund list to get new funds and delisted funds
    dump_tables_and_dfs     dump all tables and dfs of fund dm to s3
    update_stock_factor     update stock factor and factor return
    simple_check            check updating modules simply
    check_refinancing       check refinancing info
    check_refinancing_impl  check refinancing implementation info
    check_untreated_funds   check untreated new funds weekly
    pull_hedge_fund_nav     pull updated navs of hedge funds
'''


class SurfingEntrance(object):
    # TODO: rename to Surfing and move parser out of __init__
    # Make this class usable in other scripts

    def __init__(self):
        parser = argparse.ArgumentParser(description='Fund Command Tool', usage=HELP_MSG)
        parser.add_argument('command', help='Subcommand to run')
        args = parser.parse_args(sys.argv[1:2])
        if not hasattr(self, args.command):
            print("Unrecognized command")
            parser.print_help()
            exit(1)
        # use dispatch pattern to invoke method with same name
        getattr(self, args.command)()

    def _get_yesterday(self):
        yesterday = datetime.datetime.today() - datetime.timedelta(days = 1)
        yesterday_str = datetime.datetime.strftime(yesterday, '%Y%m%d')
        print(f'yesterday: {yesterday_str}')
        return yesterday_str

    def datetime_test(self):
        print(datetime.datetime.today())

    def init_raw(self):
        from .data.wrapper.mysql import RawDatabaseConnector
        from .data.view.raw_models import Base

        print('Begin to create tables for database: raw')
        Base.metadata.create_all(bind=RawDatabaseConnector().get_engine())
        print('done')

    def init_basic(self):
        from .data.wrapper.mysql import BasicDatabaseConnector
        from .data.view.basic_models import Base, FOFBase

        print('Begin to create tables for database: basic')
        Base.metadata.create_all(bind=BasicDatabaseConnector().get_engine())
        FOFBase.metadata.create_all(bind=BasicDatabaseConnector().get_engine())
        print('done')

    def init_derived(self):
        from .data.wrapper.mysql import DerivedDatabaseConnector
        from .data.view.derived_models import Base, FOFBase

        print('Begin to create tables for database: derived')
        Base.metadata.create_all(bind=DerivedDatabaseConnector().get_engine())
        FOFBase.metadata.create_all(bind=DerivedDatabaseConnector().get_engine())
        print('done')

    def init_view(self):
        from .data.wrapper.mysql import ViewDatabaseConnector
        from .data.view.view_models import Base

        print('Begin to create tables for database: view')
        Base.metadata.create_all(bind=ViewDatabaseConnector().get_engine())
        print('done')

    def init_cas(self):
        from .data.wrapper.cas.cas_init import sync_cas_table

        print('Begin to create table for database: cas')
        sync_cas_table()
        print('done')

    def download_raw(self, yesterday=None):
        if not yesterday:
            yesterday = self._get_yesterday()

        from .data.fund.raw.raw_data_downloader import RawDataDownloader
        from .util.config import SurfingConfigurator

        rq_license = SurfingConfigurator().get_license_settings('rq')
        raw_data_downloader = RawDataDownloader(rq_license)
        return raw_data_downloader.download(yesterday, yesterday), raw_data_downloader.get_updated_count()

    def update_basic(self, yesterday=None):
        if not yesterday:
            yesterday = self._get_yesterday()

        from .data.fund.basic.basic_data_processor import BasicDataProcessor

        basic_data_processor = BasicDataProcessor()
        return basic_data_processor.process_all(yesterday, yesterday), basic_data_processor.get_updated_count()

    def update_basic_not_trading_day(self, yesterday=None):
        if not yesterday:
            yesterday = self._get_yesterday()

        from .data.fund.basic.basic_data_processor import BasicDataProcessor

        basic_data_processor = BasicDataProcessor()
        return basic_data_processor.process_not_trading_day(yesterday, yesterday), basic_data_processor.get_updated_count()

    def update_derived(self, yesterday=None):
        if not yesterday:
            yesterday = self._get_yesterday()

        from .data.fund.derived.derived_data_processor import DerivedDataProcessor
        derived_data_processor = DerivedDataProcessor()
        return derived_data_processor.process_all(yesterday, yesterday), derived_data_processor.get_updated_count()

    def update_derived_not_trading_day(self, yesterday=None):
        if not yesterday:
            yesterday = self._get_yesterday()

        from .data.fund.derived.derived_data_processor import DerivedDataProcessor
        derived_data_processor = DerivedDataProcessor()
        return derived_data_processor.process_not_trading_day(yesterday, yesterday), derived_data_processor.get_updated_count()

    def update_view(self, yesterday=None):
        if not yesterday:
            yesterday = self._get_yesterday()

        from .data.fund.view.view_data_processor import ViewDataProcessor
        return ViewDataProcessor().process_all(yesterday, yesterday)

    def dev_test(self):
        from .data.manager.stock_manager import StockDataManager
        sdm = StockDataManager()
        sdm.collect()
        sdm.pickle_save()

    def simple_check(self):
        print('to check surfing configurator')
        from .util.config import SurfingConfigurator
        # SurfingConfigurator is a singleton
        _ = SurfingConfigurator(config_path='./etc/config.json')
        print('check surfing configurator done')

        print('to check raw data downloader')
        from .data.fund.raw.raw_data_downloader import RawDataDownloader
        _ = RawDataDownloader(dry_run=True)
        print('check raw data downloader done')

        print('to check basic data downloader')
        from .data.fund.basic.basic_data_processor import BasicDataProcessor
        _ = BasicDataProcessor()
        print('check basic data downloader done')

        print('to check derived data downloader')
        from .data.fund.derived.derived_data_processor import DerivedDataProcessor
        _ = DerivedDataProcessor()
        print('check derived data downloader done')

    def data_update(self):
        print('Start data update')

        yesterday = self._get_yesterday()

        failed_tasks = {}
        updated_count = {}

        print('Step 1. Start raw data download')
        raw_failed_tasks, raw_updated_count = self.download_raw(yesterday)
        if len(raw_failed_tasks) > 0:
            failed_tasks['raw'] = raw_failed_tasks
        updated_count['raw'] = raw_updated_count
        print('Step 1. Done raw data download')

        if 'em_tradedates' in raw_failed_tasks:

            print('Step 2. Not tradingday day, basic data to download')
            basic_failed_tasks, basic_updated_count = self.update_basic_not_trading_day(yesterday)
            if len(basic_failed_tasks) > 0:
                failed_tasks['basic'] = basic_failed_tasks
            updated_count['basic'] = basic_updated_count
            print('Step 2. Done update basic')

            print('Step 3. Not trading day, derived data to calculate')
            derived_failed_tasks, derived_updated_count = self.update_derived_not_trading_day(yesterday)
            if len(derived_failed_tasks) > 0:
                failed_tasks['derived'] = derived_failed_tasks
            updated_count['derived'] = derived_updated_count
            print('Step 3. Done update derived')

            print('Done data update')
            print(f'Failed tasks:\n{failed_tasks}')
            print(f'Updated count:\n{updated_count}')

            from .util.wechat_bot import WechatBot
            wechat_bot = WechatBot()
            del failed_tasks['raw'][failed_tasks['raw'].index('em_tradedates')]
            if not failed_tasks['raw']:
                del failed_tasks['raw']
            wechat_bot.send_data_update_status(yesterday, failed_tasks, updated_count)
            return

        print('Step 2. Start update basic')
        basic_failed_tasks, basic_updated_count = self.update_basic(yesterday)
        if len(basic_failed_tasks) > 0:
            failed_tasks['basic'] = basic_failed_tasks
        updated_count['basic'] = basic_updated_count
        print('Step 2. Done update basic')

        print('Step 3. Start update derived')
        derived_failed_tasks, derived_updated_count = self.update_derived(yesterday)
        if len(derived_failed_tasks) > 0:
            failed_tasks['derived'] = derived_failed_tasks
        updated_count['derived'] = derived_updated_count
        print('Step 3. Done update derived')

        print('Step 4. Start update view')
        try:
            view_failed_tasks = self.update_view(yesterday)
        except Exception:
            failed_tasks['view'] = ['fatal error']
        else:
            if len(view_failed_tasks) > 0:
                failed_tasks['view'] = view_failed_tasks
        print('Step 4. Done update view')

        print('Done data update')
        print(f'Failed tasks:\n{failed_tasks}')
        print(f'Updated count:\n{updated_count}')

        from .util.wechat_bot import WechatBot
        wechat_bot = WechatBot()
        wechat_bot.send_data_update_status(yesterday, failed_tasks, updated_count)

        print('start to check fund list')
        self.check_fund_list(yesterday, wechat_bot)
        print('check fund list done')
        print('start to dump tables and dfs')
        self.dump_tables_and_dfs(yesterday, wechat_bot)
        print('dump tables and dfs done')
        print('start to update stock factor')
        self.update_stock_factor(yesterday, wechat_bot)
        print('start to update fund factor')
        self.update_fund_factor(yesterday, wechat_bot)
        print('update stock factor done')
        print('start to check refinancing info')
        self.check_refinancing(yesterday, wechat_bot)
        self.check_refinancing_impl(yesterday, wechat_bot)
        print('check refinancing info done')
        print('start to check untreated new funds')
        self.check_untreated_funds(yesterday, wechat_bot)
        print('check untreated new funds done')

    def calc_fund_estimated_nav(self):
        # Use Jenkins to call this function every hour
        import time
        from .util.config import SurfingConfigurator
        from .data.fund.raw.raw_data_helper import RawDataHelper
        from .data.fund.basic.fund_estimated_nav_processor import FundEstimatedNavProcessor

        USE_SINA_DATA = True

        fund_estimated_nav_processor = FundEstimatedNavProcessor()
        if not fund_estimated_nav_processor.init():
            return

        if USE_SINA_DATA:
            from .data.fund.raw.sina_stock_downloader import SinaStockDownloader
            sina = SinaStockDownloader(RawDataHelper(), fund_estimated_nav_processor.stock_list)
        else:
            from .data.fund.raw.rq_raw_data_downloader import RqRawDataDownloader
            rq_license = SurfingConfigurator().get_license_settings('rq')
            rq_raw_data_downloader = RqRawDataDownloader(rq_license, RawDataHelper())

        last_time_start = None
        while True:
            try:
                time_start = datetime.datetime.now()
                # # For test only
                # if not last_time_start:
                #     time_start = datetime.datetime.fromisoformat('2020-08-06T14:59:01')

                time_start_second = time_start.second
                time_start_str = time_start.strftime('%Y-%m-%d %H:%M')
                time_start = datetime.datetime.strptime(time_start_str, '%Y-%m-%d %H:%M')

                if time_start.hour < 9 or time_start.hour == 12 or time_start.hour > 15:
                    # Exit when time_start is in range [0:00, 8:59], [12:00, 12:59] and [16:00 23:59]
                    print(f'Current time {time_start} is not trading time. Will exit.')
                    return
                elif (time_start.hour == 11 and time_start.minute >= 31) or (time_start.hour == 15 and time_start.minute >= 1):
                    # Exit when time_start is in range [11:31, 11:59] and [15:01, 15:59]
                    print(f'Current time {time_start} is not trading time. Will exit.')
                    return
                elif time_start.hour == 9 and time_start.minute < 30:
                    # At 9:00, clear fund_estimated_nav calcculated at yesterday in Cassandra
                    if time_start.minute == 29:
                        from .data.view.cas.fund_estimated_nav import FundEstimatedNav
                        from .data.wrapper.cas.cas_base import CasClient
                        CasClient().truncate_table(model=FundEstimatedNav)
                        print(f'Truncated Cassandra table {FundEstimatedNav.__table_name__}')
                    # Sleep when time_start is in range [9:00, 9:29]
                    time.sleep(max(59 - time_start_second, 1))
                elif time_start.hour == 13 and time_start.minute == 0:
                    time.sleep(max(59 - time_start_second, 1))
                elif not last_time_start or time_start.minute > last_time_start.minute or (time_start - last_time_start).seconds >= 60:
                    last_time_start = time_start
                    to_save = not (time_start.hour == 9 and time_start.minute == 30)
                    if USE_SINA_DATA:
                        stock_minute_df = sina.sina_stock_minute(to_save)
                    else:
                        stock_minute_df = rq_raw_data_downloader.stock_minute(to_save)
                    if stock_minute_df is None or stock_minute_df.empty:
                        print(f'Failed to get stock_minute_df at {time_start}')
                        continue
                    fund_estimated_nav_processor.calc_fund_estimated_nav(stock_minute_df, time_start, USE_SINA_DATA)
                    time_end = datetime.datetime.now()
                    print(f'Time for calc_fund_estimated_nav is {(time_end - time_start).total_seconds()} seconds at {time_start}')
                    if time_start.minute == 59:
                        # Exit at the end of each hour
                        return
                else:
                    time.sleep(max(59 - time_start_second, 1))
            except Exception as e:
                print(e)
                traceback.print_exc()
                return False

    def calc_index_price(self):
        # Use Jenkins to call this function every hour
        import time
        from .util.config import SurfingConfigurator
        from .data.fund.raw.raw_data_helper import RawDataHelper
        from .data.fund.raw.em_raw_data_downloader import EmRawDataDownloader
        from .data.fund.basic.realtime_index_price_processor import RealtimeIndexPriceProcessor

        realtime_index_price_processor = RealtimeIndexPriceProcessor()
        if not realtime_index_price_processor.init():
            return

        em = EmRawDataDownloader(RawDataHelper())

        last_time_start = None
        while True:
            try:
                time_start = datetime.datetime.now()
                # # For test only
                # if not last_time_start:
                #     time_start = datetime.datetime.fromisoformat('2020-08-06T14:59:01')

                time_start_second = time_start.second
                time_start_str = time_start.strftime('%Y-%m-%d %H:%M')
                time_start = datetime.datetime.strptime(time_start_str, '%Y-%m-%d %H:%M')

                if time_start.hour < 9 or time_start.hour == 12 or time_start.hour > 15:
                    # Exit when time_start is in range [0:00, 8:59], [12:00, 12:59] and [16:00 23:59]
                    print(f'Current time {time_start} is not trading time. Will exit.')
                    return
                elif (time_start.hour == 11 and time_start.minute >= 31) or (time_start.hour == 15 and time_start.minute >= 1):
                    # Exit when time_start is in range [11:31, 11:59] and [15:01, 15:59]
                    print(f'Current time {time_start} is not trading time. Will exit.')
                    return
                elif time_start.hour == 9 and time_start.minute < 30:
                    # At 9:00, clear fund_estimated_nav calcculated at yesterday in Cassandra
                    if time_start.minute == 29:
                        from .data.view.cas.realtime_index_price import RealtimeIndexPrice
                        from .data.view.cas.realtime_index_price_snapshot import RealtimeIndexPriceSnapshot
                        from .data.wrapper.cas.cas_base import CasClient
                        CasClient().truncate_table(model=RealtimeIndexPrice)
                        CasClient().truncate_table(model=RealtimeIndexPriceSnapshot)
                        print(f'Truncated Cassandra table {RealtimeIndexPrice.__table_name__}')
                    # Sleep when time_start is in range [9:00, 9:29]
                    time.sleep(max(59 - time_start_second, 1))
                elif time_start.hour == 13 and time_start.minute == 0:
                    time.sleep(max(59 - time_start_second, 1))
                elif not last_time_start or time_start.minute > last_time_start.minute or (time_start - last_time_start).seconds >= 60:
                    data_start_time = last_time_start
                    if not data_start_time:
                        data_start_time = time_start - datetime.timedelta(seconds = 60)
                    last_time_start = time_start
                    to_save = not (time_start.hour == 9 and time_start.minute == 30)
                    realtime_index_price_df = em.em_realtime_index(realtime_index_price_processor.index_ids, data_start_time, to_save)
                    if realtime_index_price_df is None or realtime_index_price_df.empty:
                        print(f'Failed to get realtime_index_price_df at {time_start}')
                        continue
                    realtime_index_price_processor.calc_realtime_index_price(realtime_index_price_df, time_start)
                    time_end = datetime.datetime.now()
                    print(f'Time for calc_realtime_index_price is {(time_end - time_start).total_seconds()} seconds at {time_start}')
                    if time_start.minute == 59:
                        # Exit at the end of each hour
                        return
                else:
                    time.sleep(max(59 - time_start_second, 1))
            except Exception as e:
                print(e)
                traceback.print_exc()
                return False

    def check_fund_list(self, yesterday=None, wechat_bot=None):
        from sqlalchemy.orm import sessionmaker
        from .data.fund.raw.raw_data_helper import RawDataHelper
        from .data.api.raw import RawDataApi
        from .data.wrapper.mysql import BasicDatabaseConnector
        from .data.view.basic_models import FundInfo

        if yesterday is None:
            yesterday = self._get_yesterday()
        if wechat_bot is None:
            from .util.wechat_bot import WechatBot
            wechat_bot = WechatBot()
        new_funds_with_name = {}
        new_delisted_funds_with_name = {}
        new_funds, new_delisted_funds = RawDataHelper.get_new_and_delisted_fund_list(yesterday)
        if new_funds is not None and new_delisted_funds is not None:
            if new_delisted_funds:
                # 如果存在新摘牌的基金 修改一下 basic.fund_info 相应基金的 end_date
                em_fund_info = RawDataApi().get_em_fund_info(new_delisted_funds)
                assert em_fund_info is not None and not em_fund_info.empty, f'failed to get em fund info for {new_delisted_funds}'
                em_fund_info = em_fund_info[['em_id', 'end_date']]
                em_fund_info = em_fund_info.set_index('em_id')

                fund_ids_succeed_to_set = []
                Session = sessionmaker(BasicDatabaseConnector().get_engine())
                db_session = Session()
                for row in db_session.query(FundInfo).filter(FundInfo.wind_id.in_(new_delisted_funds)).all():
                    try:
                        row.end_date = em_fund_info.at[row.wind_id, 'end_date']
                        fund_ids_succeed_to_set.append(row.wind_id)
                    except KeyError:
                        pass
                db_session.commit()
                db_session.close()

                fund_ids_failed_to_set = list(set(new_delisted_funds) - set(fund_ids_succeed_to_set))
                if fund_ids_failed_to_set:
                    wechat_bot.send_fund_list_that_set_end_date_failed(yesterday, fund_ids_failed_to_set)

            # 获取新增基金以及新摘牌基金列表
            agg_funds = new_funds.union(new_delisted_funds)
            if agg_funds:
                # 不是OF结尾的我们把它换成OF也去查一下
                agg_funds.update({one.split('.')[0] + '.OF' for one in agg_funds if not one.endswith('.OF')})

                api = RawDataApi()
                # 在已有的fund_info里找对应基金的信息
                wind_df = api.get_wind_fund_info(agg_funds)
                fund_name_got = {row.wind_id: row.desc_name for row in wind_df.itertuples(index=False)}
                # 在Choice的fund_info里找剩余还没找到的基金信息
                em_df = api.get_em_fund_info(agg_funds.difference(set(fund_name_got.keys())))
                fund_name_got.update({row.em_id: row.name for row in em_df.itertuples(index=False)})

                # 拆回到原来的两个列表里
                for fund_id in sorted(new_funds):
                    try:
                        new_funds_with_name[fund_id] = fund_name_got[fund_id]
                    except KeyError:
                        # 把结尾换成.OF再查一下
                        new_funds_with_name[fund_id] = fund_name_got.get(fund_id.split('.')[0] + '.OF', None)
                        if new_funds_with_name[fund_id] is not None:
                            new_funds_with_name[fund_id] += '(场内)'

                for fund_id in sorted(new_delisted_funds):
                    try:
                        new_delisted_funds_with_name[fund_id] = fund_name_got[fund_id]
                    except KeyError:
                        # 把结尾换成.OF再查一下
                        new_delisted_funds_with_name[fund_id] = fund_name_got.get(fund_id.split('.')[0] + '.OF', None)
                        if new_delisted_funds_with_name[fund_id] is not None:
                            new_delisted_funds_with_name[fund_id] += '(场内)'
        wechat_bot.send_new_fund_list(yesterday, new_funds_with_name, new_delisted_funds_with_name, RawDataHelper.get_indexes_that_become_invalid(yesterday))

    def dump_tables_and_dfs(self, yesterday=None, wechat_bot=None):
        from .data.manager.etl_tool import SynchronizerS3ForDM
        from .data.manager.manager_fund import FundDataManager

        print('to dump tables and dfs to s3')
        # 构建一个默认的DM
        try:
            fdm = FundDataManager()
            fdm.init(score_pre_calc=True, print_time=True, use_weekly_monthly_indicators=False, block_df_list=[])
        except Exception as e:
            print(f'init dm failed (err_msg){e}')

        if fdm.inited:
            # 默认DM初始化成功，用它的数据向S3同步所有df
            syncer = SynchronizerS3ForDM(dm=fdm)
            failed_dfs: List[str] = syncer.dump_all_dfs()
        else:
            print('init dm failed, dump tables only')
            syncer = SynchronizerS3ForDM()
            failed_dfs: List[str] = ['init dm']
        # 向S3同步所有table
        failed_tables: List[str] = syncer.dump_all_tables()

        # 发送同步结果
        if yesterday is None:
            yesterday = self._get_yesterday()
        if wechat_bot is None:
            from .util.wechat_bot import WechatBot
            wechat_bot = WechatBot()
        wechat_bot.send_dump_to_s3_result(yesterday, failed_tables, failed_dfs)

    def update_stock_factor(self, yesterday=None, wechat_bot=None):
        import pandas as pd
        from .stock.factor.api import StockFactorUpdater
        from .data.manager import DataManager
        from .data.api.basic import BasicDataApi

        if yesterday is None:
            yesterday = self._get_yesterday()

        # 获取下一个交易日
        trading_day_df = BasicDataApi().get_trading_day_list(start_date=yesterday)
        if trading_day_df.shape[0] <= 1:
            print(f'get trading days start with {yesterday} failed')
            failed_factors: List[str] = ['get_trading_day for stock factor updating']
        else:
            next_trading_day = trading_day_df.iloc[1, :].datetime
            print(f'got next trading day {next_trading_day}')
            end_date_dt = pd.to_datetime(yesterday, infer_datetime_format=True).date()
            next_trading_dt = pd.to_datetime(next_trading_day, infer_datetime_format=True).date()
            if end_date_dt.weekday() < next_trading_dt.weekday() and next_trading_dt < end_date_dt + datetime.timedelta(weeks=1):
                # 表明本周后边还有交易日，今天不需要更新
                print(f'stock factor only update on the last day of week, not today {end_date_dt}')
                return
            else:
                index_info = DataManager.basic_data("get_index_info")
                if index_info is not None:
                    universe_list: List[str] = index_info[index_info.is_stock_factor_universe != 0].index_id.to_list()
                    # 加进去一个default
                    universe_list.append('default')
                    print(f'[update_stock_factor] universe list: {universe_list}')

                    failed_factors: List[str] = StockFactorUpdater.do_update_everyday(yesterday, universe_list)
                else:
                    failed_factors: List[str] = ['index_info']

        if wechat_bot is None:
            from .util.wechat_bot import WechatBot
            wechat_bot = WechatBot()
        wechat_bot.send_update_factor_result(yesterday, 'stock', failed_factors)

    def update_fund_factor(self, yesterday=None, wechat_bot=None):
        import pandas as pd
        from .stock.factor.api import FundFactorUpdater
        from .data.api.basic import BasicDataApi

        if yesterday is None:
            yesterday = self._get_yesterday()

        # 获取下一个交易日
        trading_day_df = BasicDataApi().get_trading_day_list(start_date=yesterday)
        if trading_day_df.shape[0] <= 1:
            print(f'get trading days start with {yesterday} failed')
            failed_factors: List[str] = ['get_trading_day for fund factor updating']
        else:
            next_trading_day = trading_day_df.iloc[1, :].datetime
            print(f'got next trading day {next_trading_day}')
            end_date_dt = pd.to_datetime(yesterday, infer_datetime_format=True).date()
            next_trading_dt = pd.to_datetime(next_trading_day, infer_datetime_format=True).date()
            if end_date_dt.weekday() < next_trading_dt.weekday() and next_trading_dt < end_date_dt + datetime.timedelta(weeks=1):
                # 表明本周后边还有交易日，今天不需要更新
                print(f'fund factor only update on the last day of week, not today {end_date_dt}')
                return
            else:
                failed_factors: List[str] = FundFactorUpdater.do_update_every_week()

        if wechat_bot is None:
            from .util.wechat_bot import WechatBot
            wechat_bot = WechatBot()
        wechat_bot.send_update_factor_result(yesterday, 'fund', failed_factors)

    def daily_update_oversea_data(self, yesterday=None, wechat_bot=None):
        from .data.fund.raw.oversea_data_downloader import OverseaDataUpdate
        from .data.fund.derived.derived_oversea_nav_analysis import OverseaFundNavAnalysis, RawDataHelper

        if yesterday is None:
            yesterday = self._get_yesterday()

        failed_factors = []
        nav_update = OverseaDataUpdate(update_period=10, data_per_page=5)
        failed_factors.extend(nav_update.update_fund_nav())
        str_end_date = str(datetime.datetime.now().date())
        nav_analysis = OverseaFundNavAnalysis(RawDataHelper())
        nav_analysis.init(str_end_date)

        failed_factors.extend(nav_analysis.process_fund_ret())
        failed_factors.extend(nav_analysis.process_monthly_ret())
        failed_factors.extend(nav_analysis.process_period_ret())
        failed_factors.extend(nav_analysis.process_recent_ret())

        if wechat_bot is None:
            from .util.wechat_bot import WechatBot
            wechat_bot = WechatBot()
        wechat_bot.send_update_factor_result(yesterday, 'oversea', failed_factors)

    def check_refinancing(self, yesterday=None, wechat_bot=None):
        from .data.api.raw import RawDataApi

        if yesterday is None:
            yesterday = self._get_yesterday()

        df = RawDataApi().get_em_stock_refinancing_by_plan_noticed_date(yesterday, yesterday)
        if df is None or df.empty:
            return

        df = df[df.issue_object.notna()]

        if wechat_bot is None:
            from .util.wechat_bot import WechatBot
            wechat_bot = WechatBot()

        # TODO: 目前这里是hard code的
        obj_keys = {'高瓴': '^.*高瓴|天津礼仁|珠海赫成.*$'}
        for key, obj in obj_keys.items():
            df_one = df[df.issue_object.str.contains(obj)]
            if df_one.empty:
                continue
            stock_list = list(df_one.stock_id.unique())
            stock_info = RawDataApi().get_em_stock_info(stock_list=stock_list)
            if stock_info is None:
                continue

            wechat_bot.send_refinancing_key_info(yesterday, stock_info.set_index('stock_id')['name'].to_dict(), key)

    def check_refinancing_impl(self, yesterday=None, wechat_bot=None):
        from .data.api.raw import RawDataApi

        if yesterday is None:
            yesterday = self._get_yesterday()

        df = RawDataApi().get_em_stock_refinancing_impl(yesterday, yesterday)
        if df is None or df.empty:
            return

        df = df[df.issue_object.notna()]

        if wechat_bot is None:
            from .util.wechat_bot import WechatBot
            wechat_bot = WechatBot()

        # TODO: 目前这里是hard code的
        obj_keys = {'高瓴': '^.*高瓴|天津礼仁|珠海赫成.*$'}
        for key, obj in obj_keys.items():
            df_one = df[df.issue_object.str.contains(obj)]
            if df_one.empty:
                continue
            stock_list = list(df_one.stock_id.unique())
            stock_info = RawDataApi().get_em_stock_info(stock_list=stock_list)
            if stock_info is None:
                continue
            stock_price = RawDataApi().get_em_stock_price(start_date=yesterday, end_date=yesterday, stock_list=stock_list, columns=['close'])
            if stock_price is None:
                continue
            stock_info = stock_info.set_index('stock_id')[['name']].join([stock_price.drop(columns='datetime').set_index('stock_id'), df_one.set_index('stock_id').issue_price], how='outer')
            if stock_info.empty:
                continue

            wechat_bot.send_refinancing_impl_key_info(yesterday, stock_info.to_markdown(), key)

    def check_untreated_funds(self, yesterday=None, wechat_bot=None):
        import pandas as pd
        from .data.api.raw import RawDataApi
        from .data.api.basic import BasicDataApi

        if yesterday is None:
            yesterday = self._get_yesterday()

        # 获取下一个交易日
        trading_day_df = BasicDataApi().get_trading_day_list(start_date=yesterday)
        if trading_day_df.shape[0] <= 1:
            print(f'[check_untreated_funds] get trading days start with {yesterday} failed')
            return
        next_trading_day = trading_day_df.iloc[1, :].datetime
        print(f'[check_untreated_funds] got next trading day {next_trading_day}')
        end_date_dt = pd.to_datetime(yesterday, infer_datetime_format=True).date()
        next_trading_dt = pd.to_datetime(next_trading_day, infer_datetime_format=True).date()
        if end_date_dt.weekday() < next_trading_dt.weekday() and next_trading_dt < end_date_dt + datetime.timedelta(weeks=1):
            print(f'[check_untreated_funds] only check on the last day of week, not today {end_date_dt}')
            return

        fund_list = RawDataApi().get_em_fund_list(yesterday, limit=6)
        if fund_list is None:
            print(f'[check_untreated_funds] failed to get em fund list, (date){yesterday}')
            return

        latest_list = fund_list.iloc[0, :]
        latest_week = pd.to_datetime(latest_list.datetime, infer_datetime_format=True).week
        for row in fund_list.iloc[1:, :].itertuples(index=False):
            if pd.to_datetime(row.datetime, infer_datetime_format=True).week != latest_week:
                base_list = row.all_live_fund_list
                break
        else:
            print(f'[check_untreated_funds] can not find the base fund list (date){yesterday}')
            return

        untreated_new_funds = set(latest_list.all_live_fund_list.split(',')).difference(set(base_list.split(',')))
        fund_info = BasicDataApi().get_fund_info()
        untreated_new_funds = untreated_new_funds.difference(set(fund_info.wind_id.array))

        if wechat_bot is None:
            from .util.wechat_bot import WechatBot
            wechat_bot = WechatBot()

        wechat_bot.send_untreated_new_fund_list(yesterday, untreated_new_funds)

    def pull_hedge_fund_nav(self):
        from .data.manager.manager_fof_lite import FOFDataManagerLite

        MANAGER_ID = 'py1'
        FOF_ID = 'SLW695'

        fof_dm = FOFDataManagerLite()
        fof_dm.pull_hedge_fund_nav(manager_id=MANAGER_ID, fof_id=FOF_ID)


if __name__ == "__main__":
    entrance = SurfingEntrance()



import pandas as pd
import numpy as np
import math
import json
import re
import datetime
from typing import Optional, List, Union, Dict
from .base import Factor
from .calculator import Calculator
from .utils import _START_DATE
from ...data.manager import DataManager
from ...data.manager.fund_info_filter import fund_info_update, get_conv_funds, filter_fund_info, active_fund_info as get_active_fund_info
from ...data.struct import AssetTimeSpan
from ...constant import FundFactorType

class FundInfo(Factor):

    def __init__(self):
        super().__init__(f_name='FundInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self._factor = DataManager.basic_data(func_name='get_fund_info')
        self._factor = fund_info_update(self._factor).drop(columns=['_update_time'], errors='ignore')

class InMarketFundInfo(Factor):

    def __init__(self):
        super().__init__(f_name='InMarketFundInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):#TODO
        self._factor = DataManager.basic_data(func_name='get_fund_status').drop(columns=['_update_time'], errors='ignore')
        self._factor = self._factor[~self._factor['trade_status'].isnull()]
        self._factor = self._factor[~((self._factor.redem_status == '正常赎回') & (self._factor.purch_status == '正常申购'))]
        
class CloseFundInfo(Factor):

    def __init__(self):
        super().__init__(f_name='CloseFundInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self._factor = DataManager.basic_data(func_name='get_fund_open_info').drop(columns=['_update_time'], errors='ignore')

class FilteredFundInfoI(Factor):
    
    def __init__(self):
        super().__init__(f_name='FilteredFundInfoI', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(FundInfo())
        self._deps.add(InMarketFundInfo())
        self._deps.add(CloseFundInfo())

    def calc(self):
        self._factor = FundInfo().get()
        index_list = list(AssetTimeSpan.__dataclass_fields__.keys())
        self._factor = filter_fund_info(self._factor, index_list)
        in_market_fund_info = InMarketFundInfo().get()
        close_fund_info = CloseFundInfo().get()
        self._factor = self._factor[(~self._factor.fund_id.isin(in_market_fund_info.fund_id))
                                & (~self._factor.fund_id.isin(close_fund_info.fund_id))]

class TradingDay(Factor):

    def __init__(self):
        super().__init__(f_name='TradingDay', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self._factor = DataManager.basic_data(func_name='get_trading_day_list').set_index('datetime').drop(columns=['_update_time'], errors='ignore')
        self._factor = self._factor[(self._factor.index >= pd.to_datetime(_START_DATE).date())
                                 & (self._factor.index < datetime.datetime.now().date())]

class IndexInfo(Factor):

    def __init__(self):
        super().__init__(f_name='IndexInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self._factor = DataManager.basic_data(func_name='get_index_info').drop(columns=['_update_time'], errors='ignore')

class FundManagerInfo(Factor):

    def __init__(self):
        super().__init__(f_name='FundManagerInfo', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(FundInfo())

    def calc(self):
        self._factor = DataManager.derived_data(func_name='get_fund_manager_info').drop(columns=['_update_time'], errors='ignore')
        fund_info = FundInfo().get()
        self._factor = pd.merge(self._factor, fund_info[['fund_id','company_id','wind_class_2']], how='left', on='fund_id' )

class FundConvInfo(Factor):

    def __init__(self):
        super().__init__(f_name='FundConvInfo', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(FundInfo())

    def calc(self):
        fund_conv_stats = DataManager.basic_data(func_name='get_fund_conv_stats')
        fund_info = FundInfo().get()
        self._factor = get_conv_funds(fund_conv_stats, fund_info)
        self._factor = pd.DataFrame(self._factor,columns=['fund_id'])
        
class FundBenchmarkInfo(Factor):

    def __init__(self):
        super().__init__(f_name='FundBenchmarkInfo', f_type=FundFactorType.BASIC, f_level='basic')

    def calc(self):
        self._factor = DataManager.basic_data(func_name='get_fund_benchmark').drop(columns=['_update_time'], errors='ignore')

class FilteredFundInfoWhole(Factor):

    def __init__(self):
        super().__init__(f_name='FilteredFundInfoWhole', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(FundInfo())
        self._deps.add(FundBenchmarkInfo())
        self._deps.add(InMarketFundInfo())
        self._deps.add(CloseFundInfo())

    def calc(self):
        fund_info = FundInfo().get()
        benchmark_info = FundBenchmarkInfo().get()
        index_list = list(AssetTimeSpan.__dataclass_fields__.keys())
        fund_info_index = filter_fund_info(fund_info, index_list)
        fund_info_active = get_active_fund_info(fund_info, benchmark_info)
        in_market_fund_info = InMarketFundInfo().get()
        close_fund_info = CloseFundInfo().get()
        fund_info = fund_info[(fund_info.fund_id.isin(fund_info_active.fund_id))
                            | (fund_info.fund_id.isin(fund_info_index.fund_id))]
        self._factor = fund_info[(~fund_info.fund_id.isin(in_market_fund_info.fund_id))
                                & (~fund_info.fund_id.isin(close_fund_info.fund_id))]

class ActiveFundInfo(Factor): #based on fund_info fund_benchmark_info

    def __init__(self):
        super().__init__(f_name='ActiveFundInfo', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(FundInfo())
        self._deps.add(FundBenchmarkInfo())

    def calc(self):
        self._factor = get_active_fund_info(FundInfo().get(), FundBenchmarkInfo().get()).drop(columns=['_update_time'], errors='ignore')

class IndexCloseDaily(Factor):

    def __init__(self):
        super().__init__(f_name='IndexCloseDaily', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = DataManager.basic_data(func_name='get_index_price_dt', start_date=_START_DATE, columns=['close']).drop(columns=['_update_time'], errors='ignore')
        self._factor = self._factor.pivot(index='datetime',columns='index_id',values='close')
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())##

class IndexRetDaily(Factor): # based on index_close_daily

    def __init__(self):
        super().__init__(f_name='IndexRetDaily', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())
        self._deps.add(IndexCloseDaily())

    def calc(self):
        self._factor = IndexCloseDaily().get()
        self._factor = np.log(self._factor).diff(1)
        self._factor = self._factor.replace(np.Inf,np.nan).replace(-np.Inf,np.nan)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class IndexRetWeekly(Factor): # based on IndexRetDaily

    def __init__(self):
        super().__init__(f_name='IndexRetWeekly', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(IndexRetDaily())

    def calc(self):
        self._factor = IndexRetDaily().get()
        self._factor = Calculator.data_resample_weekly_ret(self._factor).dropna(axis=0, how='all')

class IndexRetMonthly(Factor): # based on IndexRetDaily

    def __init__(self):
        super().__init__(f_name='IndexRetMonthly', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(IndexRetDaily())

    def calc(self):
        self._factor = IndexRetDaily().get()
        self._factor = Calculator.data_resample_monthly_ret(self._factor).dropna(axis=0, how='all')

class FundNavDaily(Factor):

    def __init__(self):
        super().__init__(f_name='FundNavDaily', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = DataManager.basic_data(func_name='get_fund_nav_with_date', start_date=_START_DATE).drop(columns=['_update_time'], errors='ignore')
        self._factor = self._factor.pivot(index='datetime',columns='fund_id',values='adjusted_net_value')
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundRetDaily(Factor): # based on FundNavDaily

    def __init__(self):
        super().__init__(f_name='FundRetDaily', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())
        self._deps.add(FundNavDaily())

    def calc(self):
        self._factor = FundNavDaily().get()
        self._factor = np.log(self._factor).diff(1)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundSize(Factor):

    def __init__(self):
        super().__init__(f_name='FundSize', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = DataManager.basic_data(func_name='get_fund_size_range', start_date=_START_DATE).drop(columns=['_update_time'], errors='ignore')
        self._factor = self._factor.pivot(index='datetime',columns='fund_id',values='size')
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundSizeCombine(Factor): # based on FundSize

    def __init__(self):
        super().__init__(f_name='FundSizeCombine', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())
        self._deps.add(FundInfo())
        self._deps.add(FundSize())

    def calc(self):
        fund_info = FundInfo().get()
        fund_size = FundSize().get()
        fund_info['trim_name'] = fund_info.desc_name.map(lambda x : re.subn(r'[ABCDEFR]{1,2}(\(人民币\)|\(美元现汇\)|\(美元现钞\)|1|2|3)?$', '', x)[0])
        trim_name_df = fund_info.set_index('fund_id')[['trim_name']]
        fund_id_df = fund_info.set_index('trim_name')[['fund_id']]
        self._factor = fund_size.T.join(trim_name_df).groupby('trim_name').sum()
        self._factor = self._factor.join(fund_id_df).set_index('fund_id').T
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundCompanyHold(Factor):

    def __init__(self):
        super().__init__(f_name='FundCompanyHold', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = DataManager.basic_data(func_name='get_fund_company_hold', start_date=_START_DATE).drop(columns=['_update_time'], errors='ignore')
        self._factor = self._factor.pivot(index='datetime',columns='fund_id',values='institution_holds')
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundRetDailyModify(Factor): # based on fund_info fund_ret_daily

    def __init__(self):
        super().__init__(f_name='FundRetDailyModify', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())
        self._deps.add(FundInfo())
        self._deps.add(FundRetDaily())

    def calc(self):
        fund_info = FundInfo().get()
        fund_ret = FundRetDaily().get()
        _res = []
        _fund_ret_columns = fund_ret.columns.tolist()
        for fund_type, filter_ratio in Calculator.FILTER_NAV_RATIO.items():
            _info_fund_list = fund_info[fund_info['wind_class_2'].isin(Calculator.FUND_CLASSIFIER[fund_type])].fund_id.tolist()
            _fund_list = list(set(_info_fund_list).intersection(_fund_ret_columns))
            _fund_ret = fund_ret[_fund_list].copy()
            _fund_ret[_fund_ret > filter_ratio] = 0
            _res.append(_fund_ret)
        self._factor = pd.concat(_res,axis=1)
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        self._factor = self._factor.loc[:,~self._factor.columns.duplicated()].copy()

class FundInfoAsset(Factor):

    def __init__(self):
        super().__init__(f_name='FundInfoAsset', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(FundInfo())
        self._deps.add(ActiveFundInfo())
        self._deps.add(FundRetDailyModify())

    def calc(self):
        fund_info = FundInfo().get()
        active_fund_info = ActiveFundInfo().get()
        fund_ret = FundRetDailyModify().get()

        index_list = list(AssetTimeSpan.__dataclass_fields__.keys())
        fund_to_index_dict = fund_info[fund_info.index_id.isin(index_list)][['fund_id', 'index_id']].set_index('fund_id').to_dict()['index_id']
        index_fund = { index_id : [fund_idx for fund_idx, index_idx in fund_to_index_dict.items() if index_idx == index_id] for index_id in index_list}
        if 'active' in index_fund.keys():
            del index_fund['active']
        if 'conv_bond' in index_fund.keys():
            del index_fund['conv_bond']
        for index_id, fund_list in index_fund.items():
            active_list = active_fund_info[active_fund_info.index_id == index_id].fund_id
            fund_list = list(set(active_list).union(set(fund_list)).intersection(fund_ret.columns))
            index_fund[index_id] = fund_list
        fund_to_index = {}
        for index_id, fund_list in index_fund.items():
            for fund_id in fund_list:
                fund_to_index[fund_id] = index_id
        fund_info.loc[:,'bk_index_id'] = fund_info.fund_id.map(lambda x : fund_to_index.get(x))
        self._factor = fund_info

class FundRetWeeklyModify(Factor): # based on FundRetDailyModify

    def __init__(self):
        super().__init__(f_name='FundRetWeeklyModify', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(FundRetDailyModify())

    def calc(self):
        self._factor = FundRetDailyModify().get()
        self._factor = Calculator.data_resample_weekly_ret(self._factor).dropna(axis=0, how='all')

class FundRetMonthlyModify(Factor): # based on FundRetDailyModify

    def __init__(self):
        super().__init__(f_name='FundRetMonthlyModify', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(FundRetDailyModify())

    def calc(self):
        self._factor = FundRetDailyModify().get()
        self._factor = Calculator.data_resample_monthly_ret(self._factor).dropna(axis=0, how='all')

class FundNavDailyModify(Factor): # based on fund_ret_daily_modify

    def __init__(self):
        super().__init__(f_name='FundNavDailyModify', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())
        self._deps.add(FundRetDailyModify())

    def calc(self):
        self._factor = FundRetDailyModify().get()
        self._factor = np.exp(self._factor.cumsum())
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundNavWeeklyModify(Factor): # based on FundNavDailyModify

    def __init__(self):
        super().__init__(f_name='FundNavWeeklyModify', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(FundNavDailyModify())

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = Calculator.data_resample_weekly_nav(self._factor).dropna(axis=0, how='all')

class FundNavMonthlyModify(Factor): # based on FundNavDailyModify

    def __init__(self):
        super().__init__(f_name='FundNavMonthlyModify', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(FundNavDailyModify())

    def calc(self):
        self._factor = FundNavDailyModify().get()
        self._factor = Calculator.data_resample_monthly_nav(self._factor).dropna(axis=0, how='all')

class Macroeco(Factor):

    def __init__(self):
        super().__init__(f_name='Macroeco', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = DataManager.raw_data(func_name='get_em_macroeconomic_daily', codes=['EX_DR_RATIO']).drop(columns='_update_time')
        self._factor = self._factor.pivot_table(index='datetime', columns='codes', values='value')
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundBenchmarkRet(Factor): # based on fund_benchmark_info macroeco index_close_daily fund_info fund_nav_daily fund_info

    def __init__(self):
        super().__init__(f_name='FundBenchmarkRet', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())
        self._deps.add(FundBenchmarkInfo())
        self._deps.add(Macroeco())
        self._deps.add(IndexCloseDaily())
        self._deps.add(FundInfo())
        self._deps.add(FundNavDaily())

    def calc(self):
        fund_benchmark_info = FundBenchmarkInfo().get()
        macroeco = Macroeco().get()
        index_price = IndexCloseDaily().get()
        fund_info = FundInfo().get()
        fund_nav = FundNavDaily().get()
        wind_class_2_dict = fund_info.set_index('fund_id')['wind_class_2'].to_dict()

        benchmark_list: Dict[str, float] = {}
        # 遍历每一只基金的benchmark进行处理
        for row in fund_benchmark_info.itertuples(index=False):
            values: List[pd.Series] = []
            cons: float = 0
            # 空的benchmark表明我们没有对应的指数或无法解析公式
            if row.benchmark_s:
                benchmark: Dict[str, float] = json.loads(row.benchmark_s)
                benchmark_raw: Dict[str, float] = eval(row.benchmark)
                for (index, weight), index_raw in zip(benchmark.items(), benchmark_raw.keys()):
                    if index == '1':
                        # 表示benchmark中该项为常数
                        cons += weight
                    elif index in ('ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d'):
                        if weight == -1:
                            # 表示我们无法解析公式
                            # print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            break
                        else:
                            try:
                                if index == 'nonor':
                                    # 在这里我们用超额存款准备金率替代同业存款利率
                                    ra: pd.Series = macroeco.loc[:, 'EX_DR_RATIO']
                                else:
                                    ra: pd.Series = index_price.loc[:, index]
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                # print(f'[benchmark_return] Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                break
                            else:
                                values.append(ra.iloc[1:] * 0.01 * weight / Calculator.TRADING_DAYS_PER_YEAR)
                    else:
                        if weight == -1:
                            # 表示我们无法解析公式
                            # print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            break
                        else:
                            try:
                                ra: pd.Series = index_price.loc[:, index]
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                # print(f'Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                break
                            else:
                                ra = np.log(ra).diff().iloc[1:]
                                values.append(ra * weight)
                else:
                    if values or cons:
                        the_sum: float = sum(values)
                        if cons:
                            the_sum += np.log(math.pow(1 + cons, 1 / Calculator.TRADING_DAYS_PER_YEAR))
                        benchmark_list[row.fund_id] = the_sum
        self._factor =  pd.DataFrame.from_dict(benchmark_list)
        benchmark_cols = self._factor.columns.tolist()
        fund_nav_cols = fund_nav.columns.tolist()
        _res = []
        for fund_id in fund_nav_cols:
            benchmark_id = fund_id
            if not benchmark_id in benchmark_cols:    
                wind_class_2 = wind_class_2_dict[fund_id]
                if wind_class_2 not in Calculator.WIND_TYPE_DICT:
                    continue
                select_type = Calculator.WIND_TYPE_DICT[wind_class_2]
                index_price_i = index_price[Calculator.REPLACE_DICT[select_type]].rename(benchmark_id)
                benchmark_ret = np.log(index_price_i).diff(1)
                _res.append(benchmark_ret)
        self._factor = self._factor.join(pd.concat(_res,axis=1))
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())
        cols = self._factor.columns.tolist()
        cols = [i for i in cols if isinstance(i, str)]
        self._factor = self._factor[cols]

        # fund nav is null, set benchmark null
        common_columns = list(fund_nav.columns.intersection(self._factor.columns))
        con_df = fund_nav[common_columns].isnull()
        _factor = self._factor[common_columns].copy()
        _factor[con_df] = np.nan
        _cols = self._factor.columns
        _cols_not_in = [ i for i in _cols if i not in common_columns]
        self._factor = _factor.join(self._factor[_cols_not_in])

class FundBenchmarkPrice(Factor): # based on FundBenchmarkRet

    def __init__(self):
        super().__init__(f_name='FundBenchmarkPrice', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())
        self._deps.add(FundBenchmarkRet())

    def calc(self):
        self._factor = FundBenchmarkRet().get()
        self._factor = np.exp(self._factor.cumsum())
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

class FundPersonalHold(Factor):

    def __init__(self):
        super().__init__(f_name='FundPersonalHold', f_type=FundFactorType.BASIC, f_level='basic')
        self._deps.add(TradingDay())

    def calc(self):
        self._factor = DataManager.basic_data(func_name='get_history_fund_size', start_date=_START_DATE).drop(columns=['_update_time'], errors='ignore')
        self._factor = self._factor.pivot(index='datetime',columns='fund_id',values='personal_holds')
        self._factor = Calculator.data_reindex_daily_trading_day(self._factor, TradingDay().get())

'''
import surfing.data.fund_fac.basic_factors_daily as basic

attr_list = basic.__dir__()
start_idx = attr_list.index('UpdateBasicFactorStart')
update_factor_list = attr_list[start_idx+1:]

for fac in update_factor_list:
    self = eval(fac)()
    self.calc()
    self.save()
'''


import pandas as pd
import numpy as np
from functools import partial
import datetime
import time
import matplotlib.pyplot as plt
from ..resource.data_store import DataStore
from ..data.manager.manager_fund import FundDataManager
from ..data.manager.score import ScoreFunc
import dataclasses

@dataclasses.dataclass
class AssetStd:
    csi500: dict=None
    gold: dict=None
    hs300: dict=None
    mmf: dict=None
    gem: dict=None
    national_debt: dict=None
    sp500rmb: dict=None    

    # std = mean of std of each day each index
    def __post_init__(self):
        self.csi500 = self.csi500 or {'alpha':0.058899, 'fee_rate':0.003143, 'track_err':0.053827}
        self.gold = self.gold or  {'alpha':0.023139, 'fee_rate':0.003098, 'track_err':0.039580}
        self.hs300 = self.hs300 or {'alpha':0.034698, 'fee_rate':0.002682, 'track_err':0.038223}
        self.mmf = self.mmf or {'alpha':0.002942, 'fee_rate':0.000839, 'track_err':0.000626}
        self.gem = self.gem or {'alpha':0.062848, 'fee_rate':0.004320, 'track_err':0.058874}
        self.national_debt = self.national_debt or {'alpha':0.061111, 'fee_rate':0.002268, 'track_err':0.044224}
        self.sp500rmb = self.sp500rmb or {'alpha':0.012287, 'fee_rate':0.002262, 'track_err':0.009384}   

class FundIndicatorProcessor():

    TRADING_DAYS_PER_YEAR = 242
    NATURAL_DAYS_PER_YEAR = 365
    SHORT_TIME_SPAN = TRADING_DAYS_PER_YEAR
    LONG_TIME_SPAN = TRADING_DAYS_PER_YEAR * 3
    LONG_TERM_INDEX = ['hs300', 'csi500', 'mmf']
    MIN_TIME_SPAN = int(TRADING_DAYS_PER_YEAR / 4)#为了延长基金回测范围，评分最低年限3个月
    RISK_FEE_RATE = 0.025
    RISK_FEE_RATE_PER_DAY = RISK_FEE_RATE / TRADING_DAYS_PER_YEAR

    def __init__(self, dm:FundDataManager=None):
        self.dm = dm
        
    def init(self,  fund_ret_days:int=1,
                    index_id:str='hs300',
                    fund_list:list=None,
                    time_span:int=None,
                    if_min_time_span:bool=True,
                    min_time_span:int=None,
                    start_date:datetime.date=datetime.date(2015,1,1),
                    end_date:datetime.date=datetime.date(2020,6,19),
                    is_filter_c:bool=True,
                    if_normalize_devide_predecide_std:bool=False,
                    fund_id:str='673100!0',
                    func:ScoreFunc=ScoreFunc(alpha=0.3, beta=-0.6, fee_rate=-0.1),
                    if_rank_plot_only:bool=False):
        # 计算基金和大类资产指数收益考虑的天数， 默认1天
        # fund_list 计算指标基金列表 默认算资产下全部基金
        # time_span 计算指标考虑天数 默认hs300 csi500 3年，其他1年
        # min_time_span 计算指标最小天数 默认3月
        # if_min_time_span 计算指标是否天数不足，设置最小时间
        self.fund_ret_days = fund_ret_days
        
        self.time_span = time_span if time_span else self.get_time_span(index_id)
        if not if_min_time_span:
            self.min_time_span = time_span
        else:
            self.min_time_span = min_time_span if min_time_span else self.MIN_TIME_SPAN
        self.start_date = start_date
        self.end_date = end_date
        self.fund_nav = self.dm.dts.fund_nav.copy()
        self.fund_info = self.dm.dts.fund_info.copy()
        self.index_price = self.dm.dts.index_price.copy()
        self.is_filter_c = is_filter_c
        self.if_normalize_devide_predecide_std = if_normalize_devide_predecide_std
        self.fund_id = fund_id
        self.func = func if func else self.dm._score_manager.funcs.get(index_id)
        self.index_id = index_id
        self.if_rank_plot_only = if_rank_plot_only
        # 指数日线价格和基金净值 时间轴对齐 ，避免join后因为日期不存在出现价格空值
        # 指数数据历史上出现非交易日有数据，join后，基金净值出现空值，长周期时间序列算指标出空值 
        # 避免对日收益fillna(0)， 造成track error 计算不符合实际
        # 计算逻辑空值填充情况： 1. 只对基金费率空值fillna(0) 
        #                    2. manager fund 里对日线价格ffill
        #                    3. 指数数据按照基金数据重做index后ffill(有基金数据对日子,没有指数数据，比如季报出在季度末，非交易日有基金净值)
        #                    4. 有对超过基金终止日净值赋空值
        end_date = min(self.fund_nav.index[-1], self.index_price.index[-1])
        self.fund_nav = self.fund_nav.loc[:end_date]
        self.index_price = self.index_price[:end_date]
        fund_nav_index = self.fund_nav.index
        self.index_price = self.index_price.reindex(fund_nav_index).fillna(method = 'ffill')
        l1 = self.fund_nav.index.to_list().sort()
        l2 = self.index_price.index.to_list().sort()
        assert l1 == l2, f'date index of index price and fund nav are not identical, l1 {l1} lens {len(l1)}, l2 {l2} lens {len(l2)}'

        fund_to_enddate_dict = self.fund_info[['fund_id', 'end_date']].set_index('fund_id').to_dict()['end_date']
        self.fund_to_index_dict = self.fund_info[['fund_id', 'index_id']].set_index('fund_id').to_dict()['index_id']
        # 超过基金终止日的基金净值赋空
        for fund_id in self.fund_nav.columns:
            fund_end_date = fund_to_enddate_dict[fund_id]
            if self.end_date > fund_end_date:
                self.fund_nav.loc[fund_end_date:,fund_id] = np.nan

        self.fund_ret = np.log(self.fund_nav / self.fund_nav.shift(self.fund_ret_days))
        self.fund_ret = self.fund_ret.stack().reset_index(level=[0,1]).rename(columns={0:'ret'})        
        self.fund_ret['index_id'] = self.fund_ret.fund_id.apply(lambda x: self.fund_to_index_dict[x])
        self.fund_ret = self.fund_ret.pivot_table(index = ['index_id','datetime'],columns='fund_id',values='ret')
        self.index_ret = np.log(self.index_price / self.index_price.shift(self.fund_ret_days))
        self.fund_to_index_dict = {fund_id:index_id for fund_id,index_id in self.fund_to_index_dict.items() if fund_id in self.fund_ret.columns}
        self.index_list = self.fund_ret.index.levels[0]
        self.index_fund = { index_id : [fund_idx for fund_idx, index_idx in self.fund_to_index_dict.items() if index_idx == index_id] for index_id in self.index_list}
        self.start_date_dic = self.fund_info[['fund_id','start_date']].set_index('fund_id').to_dict()['start_date']
        self.fund_list = fund_list if fund_list else self.index_fund[index_id]

    def get_time_span(self, index_id):
        if index_id in self.LONG_TERM_INDEX:
            return self.LONG_TIME_SPAN
        else:
            return self.SHORT_TIME_SPAN

    def _rolling_alpha_beta(self, x, res, df):
        # 回归相关的都在这里 
        df_i = df.loc[x[0]:x[-1],]
        return self.regression_ralated_indicator(res,df_i)

    def regression_ralated_indicator(self, res, df_i):
        if sum(df_i.fund_ret) == 0:
            res.append({'alpha':np.Inf,'beta':np.Inf})
            return 1
        else:
            ploy_res = np.polyfit(y=df_i.fund_ret, x=df_i.benchmark_ret, deg=1)
            p = np.poly1d(ploy_res)
            beta = ploy_res[0]
            alpha = ploy_res[1] * 242
            res.append({'alpha': alpha, 
                        'beta': beta})
            return 1

    def indicator_calculator_item(self, fund_id:str):
        index_ret = self.index_ret[[self.index_id]].loc[self.start_date: self.end_date]
        df = self.fund_ret.loc[self.index_id][[fund_id]].loc[self.start_date: self.end_date].join(index_ret).dropna()
        df = df.rename(columns={self.index_id:'benchmark_ret', fund_id:'fund_ret'}).reset_index()
        df['year_length'] = df['datetime'].map(lambda x:(x-self.start_date_dic[fund_id]).days / self.NATURAL_DAYS_PER_YEAR)
        res = []
        pd.Series(df.index).rolling(
             window=self.time_span, min_periods=self.min_time_span).apply(
            partial(self._rolling_alpha_beta, res=res, df=df), raw=True)
        df = df.set_index('datetime')
        df = df.join(pd.DataFrame(res,index=df.index[-len(res):]))    
        df['track_err'] = (df.fund_ret - df.benchmark_ret).rolling(window=self.time_span, min_periods=self.min_time_span).std(ddof=1) * np.sqrt(self.TRADING_DAYS_PER_YEAR)
        df['fund_id'] = fund_id
        df['timespan'] = int(self.time_span / self.TRADING_DAYS_PER_YEAR)
        df['fee_rate'] = self.fund_fee[fund_id]
        return df.dropna()
        
    def indicator_calculator(self):
        # 评分无关的这里均不计算
        _tm_indicator_begin = time.time()
        self.fund_fee = self.fund_info[['fund_id','manage_fee','trustee_fee']].set_index('fund_id').fillna(0).sum(axis = 1)
        result = []
        for fund_id in self.fund_list:
            res_i = self.indicator_calculator_item(fund_id)
            result.append(res_i)
        self.fund_indicator = pd.concat(result, axis=0, sort=True).replace(np.Inf,None).replace(-np.Inf,None).dropna(subset=['beta', 'alpha','track_err']).drop(['fund_ret','benchmark_ret','timespan'], axis=1)
        _tm_indicator_end = time.time()
        print(f'calculate fund indicator with fund number {len(self.fund_list)} from {self.start_date} to {self.end_date} cost time :{round(_tm_indicator_end - _tm_indicator_begin, 2)}s')
        self.index_fund_indicator_pack = self.fund_indicator.pivot_table(index=['datetime', 'fund_id'])

    def score_calculator_item(self, dt):
        cur_d = self.index_fund_indicator_pack.loc[dt]
        fund_id_list = cur_d.index.tolist()
        fund_id_list_new = self.dm._score_manager._filter_size_and_com_hold(dt, fund_id_list, self.index_id)
        if  bool(set(fund_id_list) & set(fund_id_list_new)):
            cur_d = cur_d.reindex(fund_id_list_new).dropna()
        if cur_d.shape[0] > 1:
            if self.if_normalize_devide_predecide_std:
                for indi_i in ['alpha','fee_rate','track_err']:
                    cur_d[indi_i] = cur_d[indi_i] / AssetStd().__dict__[self.index_id][indi_i]
            nd_array = cur_d[['alpha','fee_rate','track_err']].to_numpy()
            if not self.if_normalize_devide_predecide_std:
                nd_array_calc = np.apply_along_axis(lambda x: (x - x.mean() + 1e-8)/ (x.std(ddof=1) + 1e-8), 0, nd_array)
            else:
                nd_array_calc = nd_array
            nd_array_same = cur_d[['beta']].to_numpy()
            indicator_ndarry = np.concatenate([nd_array_calc,nd_array_same], axis=1)
            normalized_indicator = pd.DataFrame(indicator_ndarry,index=cur_d.index, columns=['alpha','fee_rate','track_err','beta']).reset_index()
            normalized_indicator['datetime'] = dt
            score_list = np.apply_along_axis(lambda x: self.func.get(x), 1, indicator_ndarry)
            score_raw = pd.Series(score_list,index=cur_d.index)
            if self.index_id in self.dm._score_manager.score_penalty_param.FilterYearIndex:
                punish_funds = cur_d[cur_d.year_length < self.dm._score_manager.score_penalty_param.JudgeYearLength].index.tolist()
                for _fund_id in punish_funds:
                    score_raw[_fund_id] += self.dm._score_manager.score_penalty_param.Penalty
            score = (score_raw - score_raw.min()) / (score_raw.max() - score_raw.min())
            if self.is_filter_c:
                score = self.dm._score_manager._filter_score(score)
            score = { fund_id: s for fund_id, s in score.iteritems() }
            score_raw = { fund_id: s for fund_id, s in score_raw.iteritems() }
        else:
            score_raw = {cur_d.index[0]:1}
            score = score_raw
        return score, score_raw, normalized_indicator

    def score_calculator(self):
        _tm_score_start = time.time()
        self.score_cache = {}
        self.score_raw_cache = {}
        pad = self.fund_indicator.pivot_table(index=['datetime'])
        self.normalized_df = []
        for dt in pad.index:
            self.score_cache[dt] = {}
            self.score_raw_cache[dt] = {}
            score, score_raw, normalized_indicator = self.score_calculator_item(dt)
            self.normalized_df.append(normalized_indicator)
            self.score_cache[dt][self.index_id] = score
            self.score_raw_cache[dt][self.index_id] = score_raw
        _tm_score_end = time.time()
        self.normalized_df = pd.concat(self.normalized_df)
        print(f'calculate score cost {round(_tm_score_end - _tm_score_start, 2)}')

    def get_fund_rank(self, select_num=10):
        result = []
        score_dic = self.score_raw_cache
        fund_info = self.dm.dts.fund_info
        for d in score_dic:
            if not self.index_id in score_dic[d]:
                continue
            else:
                dic = score_dic[d][self.index_id]
                top_funds = [_[0] for _ in sorted(dic.items(),key=lambda x:x[1], reverse=True)]
                num = min(len(top_funds), select_num)
                dic = {
                    'datetime':d,
                    'rank_list': top_funds[:num]
                }            
            result.append(dic)
        df = pd.DataFrame(result).set_index('datetime')  
        rank_result = []
        for d, r in df.iterrows():
            if self.fund_id in r.rank_list:
                loc = r.rank_list.index(self.fund_id)
            else:
                loc = None
            rank_result.append({'datetime':d,'rank':loc})
        rank_df_fund = pd.DataFrame(rank_result).dropna().set_index('datetime')
        return rank_df_fund

    def get_fund_score(self):
        score_dic = self.score_raw_cache
        score_result = []
        for d in score_dic:
            if not self.index_id in score_dic[d]:
                continue
            if self.fund_id in score_dic[d][self.index_id]:
                dic = {}
                dic['datetime'] = d
                dic['score_raw'] =  score_dic[d][self.index_id][self.fund_id]
                score_result.append(dic)
        return pd.DataFrame(score_result).set_index('datetime')

    def fund_analysis(self):
        nav_df = self.dm.dts.fund_nav[[self.fund_id]].dropna()
        index_df = self.dm.dts.index_price[[self.index_id]]
        fund_indicator_fund_i = self.fund_indicator[self.fund_indicator.fund_id == self.fund_id]

        price_ratio = nav_df.join(index_df).fillna(method='ffill')
        price_ratio = price_ratio / price_ratio.iloc[0]
        price_ratio['price_ratio'] = price_ratio[self.fund_id] / price_ratio[self.index_id]
        score_df = self.get_fund_score()
        if not self.if_rank_plot_only:
            nav_df.plot.line(figsize=(12,6))       
            plt.title(f'fund nav {self.fund_id}', fontsize=17)
            plt.grid()
            price_ratio[[self.index_id]].plot.line(figsize=(12,6))
            plt.title(f'index {self.index_id} price', fontsize=17)
            plt.grid()
            price_ratio[['price_ratio']].plot.line(figsize=(12,6))
            plt.title(f'fund {self.fund_id} index {self.index_id} price ratio', fontsize=17)
            plt.grid()
            indicator_list = ['alpha','beta','track_err']
            for indicator in indicator_list:
                fund_indicator_fund_i[[indicator]].plot.line(figsize=(12,6))
                plt.title(f'fund {self.fund_id} {indicator}', fontsize=17)
                plt.grid()
            score_df.plot.line(figsize=(16,8))
            plt.title(f'fund {self.fund_id} score raw',fontsize=17)
            plt.grid()
            fund_indicator_after_normalized = self.normalized_df[self.normalized_df.fund_id == self.fund_id].copy()
            for indicator, w in self.func.__dict__.items():
                if indicator != 'beta':
                    fund_indicator_after_normalized.loc[:,f'{indicator}_component'] = abs(fund_indicator_after_normalized[indicator] * w)
                else:
                    fund_indicator_after_normalized.loc[:,f'{indicator}_component'] = abs((1 - fund_indicator_after_normalized[indicator]) * w)
            stack_df = fund_indicator_after_normalized[['alpha_component','beta_component','track_err_component','fee_rate_component']]
            stack_df = stack_df.div(stack_df.sum(axis=1).values,axis=0)
            stack_df.plot.area(figsize=(18,9),legend=False,fontsize = 17)
            plt.title('score components', fontsize=17)
            plt.legend(loc='lower left',fontsize = 14)
            plt.suptitle(str(self.func.__dict__),y=0.87,fontsize=17)
            plt.show()
        rank_df_fund = self.get_fund_rank(select_num=150)
        rank_df_fund.plot.line(figsize=(16,8))
        plt.title(f'fund {self.fund_id} rank',fontsize=17)
        plt.grid()
        plt.show()
    
    def analysis(self):
        self.indicator_calculator()
        self.score_calculator()
        self.fund_analysis()

if __name__ == "__main__":
    dm = DataStore.load_dm()
    fi = FundIndicatorProcessor(dm=dm)
    #全默认值
    fi.init(fund_ret_days=1,
            index_id='hs300',
            fund_list=None,
            time_span=None,
            if_min_time_span=True,
            min_time_span=None, 
            start_date=datetime.date(2015,1,1),
            end_date=datetime.date(2020,6,19),
            is_filter_c=True,
            if_normalize_devide_predecide_std=False,
            fund_id='673100!0',
            func=ScoreFunc(alpha=0.3, beta=-0.6, fee_rate=-0.1),
            if_rank_plot_only=False)
    fi.analysis()


import pandas as pd
import traceback
from sqlalchemy import distinct, func
from ..derived.fund_manager_processor_advanced import ManagerProcessorDev
from ...wrapper.mysql import ViewDatabaseConnector
from ...view.view_models import ManagerInfoCollection, FundManagerDailyCollection
from ...manager import DataManager
from .__init__ import mng_best_uri, mng_s3_uri

class ManagerDailyCollectionProcessor:

    def __init__(self):
        self.classfy_helper = ManagerProcessorDev

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')

    def cal(self):
        fund_manager_info = DataManager.derived_data(func_name='get_fund_manager_info')
        mng_best_fund = pd.read_parquet(mng_best_uri)
        manager_indicator = pd.read_parquet(mng_s3_uri)
        fund_info = DataManager.basic_data(func_name='get_fund_info')
        fund_manager_info = pd.merge(fund_manager_info, fund_info[['fund_id','desc_name','wind_class_1','wind_class_2']], on='fund_id')

        # 代表作逻辑较严格，对离职经理和新晋经理不评价，这里用终止日，起始日排序选出最近存续的基金作为代表作
        fund_type_dic = {}
        for fund_type, class_list in self.classfy_helper.FUND_CLASSIFIER.items():
            for class_i in class_list:
                fund_type_dic[class_i] = fund_type
        fund_manager_info.loc[:,'fund_type'] = fund_manager_info.wind_class_2.map(lambda x: fund_type_dic.get(x,'stock'))

        result = []
        last_date = mng_best_fund.index[-1]
        for fund_type in self.classfy_helper.FUND_CLASSIFIER:
            mng_info_i = fund_manager_info[fund_manager_info.fund_type == fund_type].copy()
            mng_info_i.loc[:,'best_fund'] = mng_info_i.mng_id.map(lambda x: mng_best_fund.loc[last_date,fund_type].get(x))
            no_best_fund_mng_id = mng_info_i[mng_info_i.best_fund.isnull()].mng_id.unique().tolist()
            mng_info_i = mng_info_i.set_index('mng_id')
            for mng_id in no_best_fund_mng_id:
                _df = mng_info_i.loc[[mng_id]]
                best_fund = _df.sort_values(['end_date','start_date'], ascending=[False, True]).fund_id[0]
                mng_info_i.loc[mng_id,'best_fund'] = best_fund
            mng_info_i = mng_info_i.reset_index()
            result.append(mng_info_i)
        result_df = pd.concat(result).drop(columns=['_update_time'])
        _df1 = manager_indicator.set_index(['mng_id','fund_type'])
        _df2 = result_df.set_index(['mng_id','fund_type'])[['best_fund']]
        df = _df1.join(_df2).reset_index()

        _df_all = df[df.best_fund.isnull()].copy()
        _df_existed = df[~df.best_fund.isnull()].copy()
        for r in _df_all.itertuples():
            _df = fund_manager_info[(fund_manager_info.mng_id == r.mng_id)]
            best_fund = _df.sort_values(['end_date','start_date'], ascending=[False, True]).fund_id.values[0]
            _df_all.loc[r.Index,'best_fund'] = best_fund
        df = pd.concat([_df_all,_df_existed])

        replace_dic = {
            'mdd~history_daily':'mdd_history_daily',
            'annual_ret~history_m':'annual_ret_history_m',
            'mdd~5y_m':'mdd_5y_m',
            'annual_ret~5y_m':'annual_ret_5y_m',
            'mdd~3y_m':'mdd_3y_m',
            'annual_ret~3y_m':'annual_ret_3y_m',
            'mdd~1y_m':'mdd_1y_m',
            'annual_ret~1y_m':'annual_ret_1y_m'}
        
        col_dic = {
            'mng_id':'基金经理id',
            'name':'姓名',
            'fund_type':'基金类型',
            'mdd_history_daily':'历史最大回撤',
            'annual_ret_history_m':'历史年化收益',
            'total_score':'经理总分',
            'ret_ability':'收益能力',
            'risk_ability':'抗风险能力',
            'select_time':'择时能力',
            'select_stock':'选股能力',
            'stable_ability':'稳定能力',
            'experience':'经验能力',
            'mdd_5y_m':'近五年最大回撤',
            'annual_ret_5y_m':'近五年年化收益',
            'mdd_3y_m':'近三年最大回撤',
            'annual_ret_3y_m':'近三年年化收益',
            'mdd_1y_m':'近一年最大回撤',
            'annual_ret_1y_m':'近一年年化收益',
            'trading_days':'从业天数',
            'best_fund':'代表作品',
            'company_id':'基金公司',
        }
        self.result = df.rename(columns=replace_dic)[list(col_dic.keys())]
        self.result = self.result.drop_duplicates(subset=['mng_id','fund_type'])

    def insert(self):
        self.append_data(FundManagerDailyCollection.__tablename__, self.result)

    def collection_daily(self):
        try:
            self.cal()
            self.insert()
            return True
        except Exception as e:
            print(e)
            traceback.print_exc()
            return False

    def process(self):
        failed_tasks = []
        if not self.collection_daily():
            failed_tasks.append('fund_manager_daily_collection')
        return failed_tasks

class ManagerInfoCollectionProcessor:

    def append_data(self, table_name, data_append_directly_data_df):
        if not data_append_directly_data_df.empty:
            with ViewDatabaseConnector().managed_session() as mn_session:
                try:
                    mn_session.execute(f'TRUNCATE TABLE {table_name}')
                    mn_session.commit()
                except Exception as e:
                    print(f'Failed to truncate table {table_name} <err_msg> {e}')
            data_append_directly_data_df.to_sql(table_name, ViewDatabaseConnector().get_engine(), index = False, if_exists = 'append')
            print('新数据已插入')
        else:
            print('没有需要插入的新数据')

    def collection_daily(self):
        mng_info = [
            {
                'mng_id':'16E872',
                'mng_name':'萧楠',
                'introduction':'消费之王，坚守价值投资，屡获奖项',
                'fund_type': 'stock',
            },
            {
                'mng_id':'170F31',
                'mng_name':'谢治宇',
                'introduction':'自下而上精选个股，行业均衡，成长投资，代表作7年6倍',
                'fund_type': 'stock',
            },
            {
                'mng_id':'7CABBB',
                'mng_name':'袁芳',
                'introduction':'擅长行业消费+科技，风格均衡，高换手分散持股',
                'fund_type': 'stock',
            },
            {
                'mng_id':'17109A',
                'mng_name':'曲扬',
                'introduction':'投资风格以稳健著称，获得第七届英华奖五年期股票投资最佳基金经理奖项',
                'fund_type': 'stock',
            },
            {
                'mng_id':'16E885',
                'mng_name':'张坤',
                'introduction':'收益优秀，8年6倍，700亿规模',
                'fund_type': 'stock',
            },
            {
                'mng_id':'7C3A95',
                'mng_name':'杨浩',
                'introduction':'TMT行业有着深入的研究，交银三剑客',
                'fund_type': 'stock',
            },
            {
                'mng_id':'7B55CF',
                'mng_name':'谭小兵',
                'introduction':'任职四年半回报230%',
                'fund_type': 'stock',
            },
            {
                'mng_id':'18377A',
                'mng_name':'王君正',
                'introduction':'通过稳健的投资框架，获得过金牛奖、明星基金奖和金基金奖共6座，业内少有的大满贯基金经理',
                'fund_type': 'stock',
            },
            {
                'mng_id':'7BE82B',
                'mng_name':'何帅',
                'introduction':'绝对收益；成长价值；性价比为王；长期优异；交银三剑客',
                'fund_type': 'stock',
            },
            {
                'mng_id':'78F3D1',
                'mng_name':'孙伟',
                'introduction':'连续5年金牛奖，全市场获此荣誉的基金经理仅3人三次获得明星基金奖',
                'fund_type': 'stock',
            },
            {
                'mng_id':'19A22E',
                'mng_name':'张清华',
                'introduction':'股债双修研究深入，择时能力突出',
                'fund_type': 'bond',
            },
            {
                'mng_id':'01109',
                'mng_name':'桂跃强',
                'introduction':'股债配置能力强，投资权益的风格就属于稳健型，擅长使用权益类资产进行稳定增强',
                'fund_type': 'bond',
            },
            {
                'mng_id':'4F9879',
                'mng_name':'姚秋',
                'introduction':'股债全能的投资界“名将”，宏观层面把握机会，追求长期稳健回报资产配置能力出色',
                'fund_type': 'bond',
            },
            {
                'mng_id':'4F9865',
                'mng_name':'张雅君',
                'introduction':'金牛债券基金经理，同类产品连续六年正收益，金牛基金、明星基金，五星债券基金',
                'fund_type': 'bond',
            },
            {
                'mng_id':'C3B3E7',
                'mng_name':'蒋利娟',
                'introduction':'第六届中国基金业英华奖三年期纯债投资最佳基金经理、泰康资产公募事业部固收投资负责人',
                'fund_type': 'bond',
            },
            {
                'mng_id':'00707',
                'mng_name':'苏玉平',
                'introduction':'五年期二级债最佳基金经理奖”，连续三年在英华奖评选中榜上有名',
                'fund_type': 'bond',
            },
            {
                'mng_id':'17036A',
                'mng_name':'于泽雨',
                'introduction':'2019年金牛奖最佳人气金牛基金经理得主、擅长分析基本面 擅长长周期操作',
                'fund_type': 'bond',
            },
            {
                'mng_id':'7A8375',
                'mng_name':'张雪',
                'introduction':'擅长使用久期择时，杠杆择时，注重风险调整后收益，夏普比率2.58',
                'fund_type': 'bond',
            },
            {
                'mng_id':'B9F063',
                'mng_name':'邬传雁',
                'introduction':'现任泓德基金副总经理。风格稳健型，股债研究均深入，擅长股债择时，权益类增强',
                'fund_type': 'bond',
            },
            {
                'mng_id':'00524',
                'mng_name':'李建',
                'introduction':'擅长股债择时和大类资产配置，严控回撤同时实现长期正收益的灵活配置型代表',
                'fund_type': 'stock',
            },
            {
                'mng_id':'7BEBB3',
                'mng_name':'刘博',
                'introduction':'选股能力出色，风险敞口低，近两年表现出色，任职年化收益45.39%。',
                'fund_type': 'stock',
            },
            {
                'mng_id':'16094C',
                'mng_name':'庄园',
                'introduction':'灵活配置股债，回撤控制和收益能力出色，获债券型优胜金牛基金等高含金量奖项',
                'fund_type': 'stock',
            },
            {
                'mng_id':'164E99',
                'mng_name':'宋炳珅',
                'introduction':'8次获得金牛奖，代表作年化收益7.63%，回撤控制在5.05%水平。工银瑞信“固收+”王牌团队成员',
                'fund_type': 'bond',
            },
        ]        
        self.result = pd.DataFrame(mng_info)

    def insert(self):
        self.append_data(ManagerInfoCollection.__tablename__, self.result)


from http.server import HTTPServer
from http.server import BaseHTTPRequestHandler
from urllib.parse import urlparse, quote, unquote
from xml.etree import ElementTree
import random
import threading
# import ssl
from ...util.config import Configurator
from ...util.struct import WxMsgItem
from ...util import Singleton
from ...util.wx_crypto.WXBizMsgCrypt3 import WXBizMsgCrypt

class MessageCrypt(metaclass=Singleton):

    def __init__(self):
        token = Configurator().config['message_receiver']['wechat']['token']
        aes_key = Configurator().config['message_receiver']['wechat']['aes_key']
        receive_id = Configurator().config['message_receiver']['wechat']['app_id']
        self._msg_crypt = WXBizMsgCrypt(token, aes_key, receive_id)
    
    def get_msg_crypt(self):
        return self._msg_crypt

class WxHTTPRequestHandler(BaseHTTPRequestHandler):

    def do_GET(self):
        query = urlparse(self.path).query
        query_components = dict(qc.split("=") for qc in query.split("&"))
        if ("signature" not in query_components or "timestamp" not in query_components or 
            "nonce" not in query_components or "echostr" not in query_components):
            self.send_response(400)
            self.send_header('Content-Type', 'text/plain; charset=utf-8')
            self.end_headers()

        signature = unquote(query_components["signature"])
        timestamp = unquote(query_components["timestamp"])
        nonce = unquote(query_components["nonce"])
        echostr = unquote(query_components["echostr"])

        ret, rsp_echo = MessageCrypt().get_msg_crypt().WechatVerifyURL(signature, timestamp, nonce, echostr)
        print(f'{ret} {rsp_echo}')
        self.send_response(200)
        self.send_header('Content-Type', 'text/plain; charset=utf-8')
        self.end_headers()
        self.wfile.write(rsp_echo)

    def do_POST(self):
        query = urlparse(self.path).query
        query_components = dict(qc.split("=") for qc in query.split("&"))
        if ("signature" not in query_components or "timestamp" not in query_components or 
            "nonce" not in query_components):
            self.send_response(400)
            self.send_header('Content-Type', 'text/plain; charset=utf-8')
            self.end_headers()

        signature = unquote(query_components["signature"])
        timestamp = unquote(query_components["timestamp"])
        nonce = unquote(query_components["nonce"])

        # Get the size of post data
        content_length = int(self.headers['Content-Length'])
        # Get the post data
        post_data = self.rfile.read(content_length)

        ret, xml_content = MessageCrypt().get_msg_crypt().WechatDecryptMsg(post_data, signature, timestamp, nonce)
        xml_root = ElementTree.fromstring(xml_content)
        from_user = xml_root.find('FromUserName').text
        to_user = xml_root.find('ToUserName').text
        msg_type = xml_root.find('MsgType').text
        print(xml_content)

        if msg_type == 'text':
            content = xml_root.find('Content').text
            msg_id = xml_root.find('MsgId').text
            data = {
                'msgid': msg_id,
                'tolist': [from_user],
                'msgtype': msg_type,
                'msgtime': int(timestamp) * 1000, # convert to ms
                'text': {'content': content}
            }
            msg = WxMsgItem(data)
            print(msg)

            with self.server.msg_mutex:
                self.server.msgs.append(msg)
        elif msg_type == 'event':
            # When a user scans QR on page of MyProfile/WeChat
            # https://developers.weixin.qq.com/doc/offiaccount/Message_Management/Receiving_event_pushes.html
            event = xml_root.find('Event').text.lower()
            if event == 'subscribe' and xml_root.find('EventKey') is not None and xml_root.find('EventKey').text:
                # When an new user scans QR and follows IFA
                event_key = xml_root.find('EventKey').text
                # "qrscene_" in front of scene_id
                scene_id = event_key[8:]
                print(scene_id)
            elif event == 'scan':
                # When an existing user scans QR
                scene_id = xml_root.find('EventKey').text
                print(scene_id)

        self.send_response(200)
        self.send_header('Content-Type', 'text/plain; charset=utf-8')
        self.end_headers()

class MyHTTPServer(HTTPServer):
    def __init__(self, server_address, RequestHandlerClass):
        super().__init__(server_address, RequestHandlerClass)
        self.msg_mutex = threading.Lock()
        self.msgs = []

    def get_msg(self, count):
        with self.msg_mutex:
            result = self.msgs[:count]
            self.msgs = self.msgs[count:]
        return result

class MessageReceiver(object):
    def __init__(self):
        # keyfile = Configurator().config['cert']['keyfile']
        # certfile = Configurator().config['cert']['certfile']

        self.server = MyHTTPServer(('0.0.0.0', 80), WxHTTPRequestHandler)
        # server.socket = ssl.wrap_socket(server.socket, 
        #     keyfile=keyfile, 
        #     certfile=certfile, server_side=True)

    def run(self):
        self.server.serve_forever()

    def stop(self):
        self.server.shutdown()

    def get_msg(self, count):
        return self.server.get_msg(count)

if __name__ == '__main__':
    mr = MessageReceiver()
    print('Starting server, use <Ctrl-C> to stop')
    mr.run()



from typing import Dict

import numpy as np
import pandas as pd

from .factor_types import MomentumFactor, MixFactorBase
from .basic_factors import StockPostPriceFactor
from .utils import get_day


class AdjMomentumDateFactor(MomentumFactor):
    # 动量（X月）
    def __init__(self, num):
        super().__init__('m'+str(num)+'d')
        self.num = num
        self._deps.add(StockPostPriceFactor())

    def calc(self):
        spp = StockPostPriceFactor().get()
        ls = spp.index  # trading_day_list
        fac = spp.copy()
        for i in range(0, len(spp)):
            try:
                temp = get_day(30, ls, spp.index[i])
                temp2 = get_day(self.num, ls, spp.index[i])
                fac.iloc[i] = spp.loc[temp] / spp.loc[temp2] - 1
            except Exception:
                fac.iloc[i] = np.nan
        self._factor = fac


class AdjMomentum3MFactor(AdjMomentumDateFactor):
    # 动量（3月）
    def __init__(self):
        super().__init__(90)


class AdjMomentum6MFactor(AdjMomentumDateFactor):
    # 动量（6月）
    def __init__(self):
        super().__init__(180)


class AdjMomentum12MFactor(AdjMomentumDateFactor):
    # 动量（12月）
    def __init__(self):
        super().__init__(365)


class MixMomentumFactor(MixFactorBase, MomentumFactor):
    # 合成因子
    def __init__(self):
        super(MixFactorBase, self).__init__('mixm')
        self._factor: Dict[pd.DataFrame] = {}
        self._deps.add(AdjMomentum3MFactor())
        self._deps.add(AdjMomentum6MFactor())
        self._deps.add(AdjMomentum12MFactor())

    def calc(self, universe: str):
        self._factor[universe] = AdjMomentum3MFactor().get_normalized(universe) + AdjMomentum6MFactor().get_normalized(universe) + AdjMomentum12MFactor().get_normalized(universe)



import os
import datetime
import numpy as np
import pandas as pd
from scipy.stats import pearsonr
from ..structure import PositionInfo
from ..data.wrapper import DataWrapper

class BacktestReport:
    # for single strategy

    def __init__(self,backtest_result=None,period=None,date_list=None,price_data=None,factor_data=None,stock_pool_data=None,stock_info=None,base_path=None,common_pool=None,pool=None):
        self.backtest_result = backtest_result
        self.hold_period     = period
        self.date_list       = date_list
        self.price_data      = price_data
        self.factor_data     = factor_data
        self.stock_info      = stock_info
        self.stock_pool_data = stock_pool_data
        self.base_path       = base_path
        self.common_pool     = common_pool
        self.pool            = pool
        self.post_open       = self.load_price('post_open')
        self.post_close      = self.load_price('post_close')
        self.post_vwap       = self.load_price('post_vwap')
        self.index_close     = self.load_index()
        self.report          = {} 

    # data process part
    def get_df(self, dic):
        return pd.DataFrame(data = dic['data'], columns = dic['columns'] )
    
    def get_sell_day(self, date):
        idx = self.date_list.index(date) + self.hold_period
        return self.date_list[idx]

    def filter_by_pool_2_series(self, dic, pool):
        return  pd.Series({k: v for k, v in  dic.items() if k in pool}).sort_index(ascending=True)

    def filter_dict_by_key(self, dic):
        dic = dict(filter(lambda elem: (elem[0] >= self.date_list[0] and elem[0] <= self.date_list[-1]), dic.items()))
        return dic

    def load_price(self, price_type):
        paths = os.path.join(self.base_path, price_type)
        meta_info = DataWrapper(paths).meta_info
        columns = meta_info['tickers']+['time']
        df = DataWrapper.base_load_npz(paths, price_type, columns)
        df = df.fillna(method='ffill')
        df.index = df.time.dt.strftime('%Y-%m-%d')
        df.drop(columns=['time'],inplace=True)
        df = df[self.common_pool]
        df = df.astype(float).round(3).copy()
        df['time'] = df.index
        data_dic = df.to_dict('index')
        return self.filter_dict_by_key(data_dic)

    def load_index(self):
        price_type = 'index_close'
        paths = os.path.join(self.base_path, price_type)
        meta_info = DataWrapper(paths).meta_info
        columns = meta_info['tickers']+['time']
        df = DataWrapper.base_load_npz(paths, price_type, columns)
        df = df[[self.pool, 'time']]
        df = df.loc[df['time'] >= self.backtest_result['mv']['data'][0][0]].copy()
        df = df.loc[df['time'] <= self.date_list[-1]].copy()
        return df.reset_index(drop=True)

    def analysis_df(self, df):
        df = pd.DataFrame(df)
        r1 = (df['close>open'].sum() /  df['close>open'].shape[0]) * 100
        r1 = round(r1, 2)                               #close > open rate
        r2 = (df['vwap>open'].sum() /  df['vwap>open'].shape[0]) * 100
        r2 = round(r2, 2)                               #vwap > open rate
        r3 = (df['close>vwap'].sum() /  df['close>vwap'].shape[0]) * 100
        r3 = round(r3, 2)                               #close > vwap  rate
        return r1, r2, r3
       
    # calculate part
    def result_mv(self):
        self.mv_df = self.get_df(self.backtest_result['mv'])    
        self.mv_result = {}
        self.mv_result['index_nv'] = self.index_close[self.pool] / self.index_close[self.pool].values[0]
        self.mv_result['strag_nv'] = self.mv_df['mv'] / self.mv_df['mv'].values[0]
        self.mv_result['hedge_nv'] = 0.5 * self.mv_result['index_nv'] + 0.5 * self.mv_result['strag_nv']
        self.mv_result['index_nv'] = self.mv_result['index_nv'].tolist()
        self.mv_result['strag_nv'] = self.mv_result['strag_nv'].tolist()
        self.mv_result['hedge_nv'] = self.mv_result['hedge_nv'].tolist()
        self.mv_result['date']     = self.mv_df['date'].tolist()

    def result_basic(self):
        st_b_day = datetime.datetime.strptime(self.mv_df.date.iloc[ 0], '%Y-%m-%d')
        st_e_day = datetime.datetime.strptime(self.mv_df.date.iloc[-1], '%Y-%m-%d')
        self.years =  (st_e_day - st_b_day).days / 365
        self.total_return = self.mv_df.mv.iloc[-1] / self.mv_df.mv.iloc[0]
        self.annualized_return = np.exp(np.log(self.total_return)/self.years) - 1
        self.annualized_volatiltity = (self.mv_df.mv.shift(1) / self.mv_df.mv).std() * np.sqrt((self.mv_df.shape[0] - 1) / self.years)
        self.sharpe = self.annualized_return / self.annualized_volatiltity
        
        self.buys_fee = PositionInfo.FEE_RATE['STOCK']['BUYFEE']
        self.sell_fee = PositionInfo.FEE_RATE['STOCK']['SELLFEE']
        self.slippage = PositionInfo.FEE_RATE['STOCK']['SLIPPAGE']

    def result_mdd(self):
        mdd_part1 = (self.mv_df.loc[:, 'mv'] / self.mv_df.loc[:, 'mv'].rolling(10000, min_periods=1).max())
        self.mdd_date1 = self.mv_df.loc[self.mv_df.loc[:mdd_part1.idxmin(),'mv'].idxmax(),'date']
        self.mdd_date2 = self.mv_df.loc[mdd_part1.idxmin(),'date']
        self.mdd = 1 - mdd_part1.min()

        self.mdd_below_level = 1 - self.mv_df.loc[:, 'mv'].min() / self.mv_df.loc[0, 'mv']
        if self.mdd_below_level  > 0:
            date_idx  = np.where(self.mv_df.loc[:,'mv'] <= self.mv_df.loc[0, 'mv'])[0][0]
            self.mdd_below_level_date1 = self.mv_df.loc[date_idx, 'date']
            self.mdd_below_level_date2 = self.mv_df.loc[self.mv_df.loc[:, 'mv'].idxmin(), 'date']
        else:
            self.mdd_below_level_date1 = None
            self.mdd_below_level_date2 = None

    def result_ic(self):
        self.ic_result = []
        for date in self.date_list[:-self.hold_period]:
            sel_date = self.get_sell_day(date)
            factor_1  = self.filter_by_pool_2_series(self.factor_data[date], self.stock_pool_data[date])
            price_1   = self.filter_by_pool_2_series(self.price_data[date],  self.stock_pool_data[date])
            price_2   = self.filter_by_pool_2_series(self.price_data[sel_date],  self.stock_pool_data[date])
            ic = pearsonr(factor_1.rank(pct=True), (price_2 / price_1).rank(pct=True))[0]
            self.ic_result.append({'date':date,'ic':ic})
        
        ic_df = pd.DataFrame(self.ic_result)
        self.ic_data = {
            'ic'    :   [0] + ic_df['ic'].values.tolist() + [0] * self.hold_period,
        }
        self.mean_ic = ic_df['ic'].mean()
        self.std_ic = ic_df['ic'].std()
        self.ir = self.mean_ic / self.std_ic

    def result_trade_pair(self):    
        trade_list = self.get_df(self.backtest_result['trade']).to_dict('records')
        trade_list_b = [_ for _ in trade_list if _['direction'] == 'BUY']
        trade_list_s = [_ for _ in trade_list if _['direction'] == 'SELL']
        self.trade_pair = []
        for trade_b in trade_list_b:
            for trade_s in trade_list_s:
                if trade_s['trade_volume'] == trade_b['trade_volume']: 
                    pct = (trade_s['trade_price'] / trade_b['trade_price'] - 1) * 100
                    self.trade_pair.append({'buy':trade_b, 'sel':trade_s, 'pct': pct})
                    break
        self.trade_pair = sorted(self.trade_pair, key = lambda i: i['pct'], reverse=True) 
        self.trade_pair_top = []
        self.trade_pair_bot = []
        for i in self.trade_pair[:10]:
            self.trade_pair_top.append(
                {
                    'pct'       :   str(round(i['pct'],2)) + '%',
                    'ticker'    :   i['buy']['ticker'],
                    'buy_time'  :   i['buy']['trade_time'],    
                    'sel_time'  :   i['sel']['trade_time'],
                    'chn_name'  :   self.stock_info[i['buy']['ticker']]['chn_name'],
                }
            ) 
        for i in self.trade_pair[-10:]:
            self.trade_pair_bot.append(
                {
                    'pct'       :   str(round(i['pct'],2)) + '%',
                    'ticker'    :   i['buy']['ticker'],
                    'buy_time'  :   i['buy']['trade_time'],    
                    'sel_time'  :   i['sel']['trade_time'],
                    'chn_name'  :   self.stock_info[i['buy']['ticker']]['chn_name'],
                }
            ) 

        self.mean_ret = np.mean([ _['pct'] for _ in self.trade_pair ])
        ticker_list = list(set([ _['ticker'] for _ in trade_list ]))
        ret_result = []
        recent_buy = []
        for ticker in ticker_list:
            ret_ticker = 0
            trade_time = 0
            trade_ticker_list = [ _ for _ in trade_list if _['ticker'] == ticker ]
            for i in trade_ticker_list:
                if (trade_ticker_list.index(i) == len(trade_ticker_list) - 1) and i['direction'] == 'BUY':
                    recent_buy.append({'ticker': ticker, 'date' : i['trade_time']})
                    continue
                trade_time += 1
                ret_ticker += i['trade_price'] * i['trade_volume'] * -1 if i['direction'] == 'BUY' else i['trade_price'] * i['trade_volume']
            ret_result.append({'ticker': ticker, 'earn_amount': ret_ticker, 'buy_times': int(trade_time/2)})
        
        self.stock_ret = pd.DataFrame(ret_result)
        ret_result = []
        self.stock_ret = self.stock_ret.sort_values(by='earn_amount', ascending=False)
        self.stock_ret = self.stock_ret.reset_index(drop=True)
        chn_name_list    = []
        sw_industry_list = []
        for ticker in self.stock_ret.ticker:
            chn_name_list.append(self.stock_info[ticker]['chn_name'])
            sw_industry_list.append(self.stock_info[ticker]['shenwan_industry_name'])

        self.stock_ret['chn_name'] =  chn_name_list
        self.stock_ret['sw_industry'] = sw_industry_list

        self.stock_ret['earn_amount'] = self.stock_ret['earn_amount'].astype(int)


        self.recent_buy = pd.DataFrame(recent_buy)
        chn_name_list    = []
        sw_industry_list = []
        for ticker in self.recent_buy.ticker:
            chn_name_list.append(self.stock_info[ticker]['chn_name'])
            sw_industry_list.append(self.stock_info[ticker]['shenwan_industry_name'])
        self.recent_buy['chn_name'] =  chn_name_list
        self.recent_buy['sw_industry'] = sw_industry_list
        self.recent_buy = self.recent_buy.to_dict('records')

        self.stock_ret_top10     = self.stock_ret.head(10).to_dict('records')
        self.stock_ret_bot10     = self.stock_ret.tail(10).to_dict('records')
        self.stock_ret_barchart  = {
            'earn_amount'   :  self.stock_ret['earn_amount'].values.tolist(),
            'ticker'        :  self.stock_ret['ticker'].values.tolist(),
        }
        
    def result_switch_price(self):
        buy_date_result = []
        sel_date_result = []
        for trade_i in self.backtest_result['trade']['data']:
            ticker      = trade_i[1]
            direction   = trade_i[2]
            date        = trade_i[5]
            open_p = self.post_open[date][ticker]
            cloz_p = self.post_close[date][ticker]
            vwap_p = self.post_vwap[date][ticker]
            con1 = (cloz_p > open_p) * 1
            con2 = (vwap_p > open_p) * 1
            con3 = (cloz_p > vwap_p) * 1
            if direction == 'BUY':
                buy_date_result.append({'close>open':con1,
                                        'vwap>open':con2,
                                        'close>vwap':con3})
            else:
                sel_date_result.append({'close>open':con1,
                                        'vwap>open':con2,
                                        'close>vwap':con3})
        self.buy_close_open, self.buy_vwap_open, self.buy_close_vwap = self.analysis_df(buy_date_result)
        self.sel_close_open, self.sel_vwap_open, self.sel_close_vwap = self.analysis_df(sel_date_result)
        

    def generate_report(self): 
        self.report['period']                   =   self.hold_period     
        self.report['year']                     =   str(round(self.years,2))
        self.report['annualized_return']        =   str(round(self.annualized_return, 2) * 100) + '%'
        self.report['annualized_volatility']    =   str(round(self.annualized_volatiltity,2))
        self.report['sharpe']                   =   str(round(self.sharpe,2))
        self.report['buy_fee']                  =   str(self.buys_fee)
        self.report['sell_fee']                 =   str(self.sell_fee)
        self.report['slippage']                 =   str(self.slippage)
        self.report['mdd']                      =   str(round(self.mdd*100,2)) + '%'
        self.report['mdd_date1']                =   self.mdd_date1
        self.report['mdd_date2']                =   self.mdd_date2
        self.report['mdd_below_level']          =   str(round(self.mdd_below_level*100,2)) + '%'
        self.report['mdd_below_level_date1']    =   self.mdd_below_level_date1
        self.report['mdd_below_level_date2']    =   self.mdd_below_level_date2
        self.report['ic_data']                  =   self.ic_data 
        self.report['mean_ic']                  =   str(round(self.mean_ic,4))
        self.report['ir']                       =   str(round(self.ir,4))
        self.report['buy_date_close>open']      =   str(self.buy_close_open) + '%'
        self.report['buy_date_vwap>open']       =   str(self.buy_vwap_open) + '%'
        self.report['buy_date_close>vwap']      =   str(self.buy_close_vwap) + '%'
        self.report['sel_date_close>open']      =   str(self.sel_close_open) + '%'
        self.report['sel_date_vwap>open']       =   str(self.sel_vwap_open) + '%'
        self.report['sel_date_close>vwap']      =   str(self.sel_close_vwap) + '%'
        self.report['trades_mean_ret']          =   str(round(self.mean_ret,3)) + '%'
        self.report['total_trade_fee']          =   str(self.buys_fee + self.sell_fee + 2 * self.slippage)
        self.report['trade_pair_top']           =   self.trade_pair_top
        self.report['trade_pair_bot']           =   self.trade_pair_bot
        self.report['stock_ret_top10']          =   self.stock_ret_top10
        self.report['stock_ret_bot10']          =   self.stock_ret_bot10
        self.report['recent_buy']               =   self.recent_buy
        self.report['stock_ret_barchart']       =   self.stock_ret_barchart
        self.report['mv']                       =   self.mv_result


    # main function of calculate backtest result
    def analysis_backtest_result(self):
        self.result_mv()
        self.result_basic()
        self.result_mdd()
        self.result_ic()
        self.result_trade_pair()
        self.result_switch_price()
        self.generate_report()
        return self.report


from typing import List, Dict
import pandas as pd
import numpy as np
import math
import datetime
import traceback
import json
import statsmodels.api as sm
import time
from multiprocessing import Pool
from scipy.optimize import Bounds, minimize
import statsmodels.api as sm
from statsmodels import regression
from statsmodels.tsa.ar_model import AutoReg
from pandas.tseries.offsets import DateOffset
from ...api.raw import RawDataApi
from ...api.basic import BasicDataApi
from ...api.derived import DerivedDataApi
from ...view.derived_models import FundIndicatorGroup
from .derived_data_helper import DerivedDataHelper, normalize, score_rescale


class FundIndicatorProcessorGroup:

    TRADING_DAYS_PER_YEAR = 242
    TRADING_DAYS_PER_MONTH = 20
    RISK_FEE_RATE = 0.011
    RISK_FEE_RATE_PER_DAY = RISK_FEE_RATE / TRADING_DAYS_PER_YEAR
    NATURAL_DAYS_PER_YEAR = 360
    REPORT_DATE_LIST = ['0331', '0630', '0930', '1231']
    FUND_CLASSIFIER = {
        'QDII': ['国际(QDII)股票型基金', '国际(QDII)债券型基金', '国际(QDII)另类投资基金', '国际(QDII)混合型基金', ],
        'stock': ['普通股票型基金', '偏股混合型基金',  '被动指数型基金', '增强指数型基金', '平衡混合型基金', '灵活配置型基金', '股票多空'], 
        'bond': ['中长期纯债型基金', '短期纯债型基金', '偏债混合型基金', '增强指数型债券基金', '被动指数型债券基金','混合债券型二级基金', '混合债券型一级基金'],
        'mmf': ['货币市场型基金'],
        'index': ['被动指数型基金', '增强指数型基金', '增强指数型债券基金', '被动指数型债券基金', '商品型基金', 'REITs'],
    }
    REPLACE_DICT = {'stock': 'csi_stockfund',
                     'bond': 'csi_boodfund', 
                     'mmf':  'mmf', 
                     'QDII': 'csi_f_qdii', 
                     'index':'hs300'}

    def __init__(self, data_helper: DerivedDataHelper):
        self.data_helper = data_helper
        self.basic_api = BasicDataApi()

    def init(self, end_date: str = '20201023'):
        self.end_date = pd.to_datetime(end_date, infer_datetime_format=True).date()
        self.wind_type_dict = {}
        for type_i, type_list in self.FUND_CLASSIFIER.items():
            for _ in type_list:
                self.wind_type_dict.update({_:type_i})
        
        # 交易日
        self.trading_days = self.basic_api.get_trading_day_list().drop(columns='_update_time').set_index('datetime')
        self.trading_days = self.trading_days.loc[: self.end_date]
        self.begin_date = self.trading_days.tail(242).index[0]
        # 基金信息
        self.fund_info = self.basic_api.get_fund_info().drop(columns='_update_time')
        self.fund_info = self.fund_info[(self.fund_info.end_date >= self.end_date) 
                                       & (self.fund_info.structure_type <= 1)
                                       & (~self.fund_info.wind_class_2.isnull())
                                       & (self.fund_info.start_date <= self.begin_date)]
        self.wind_class_2_dict = self.fund_info.set_index('fund_id')['wind_class_2'].to_dict()

        # 成立半年不算 TODO
        # 基金净值
        fund_list = self.fund_info.fund_id.tolist()
        #fund_list = ['000001!0','000003!0','000004!0','000005!0','000006!0','000008!0','000009!0','000010!0']  # if test
        self.fund_nav = self.basic_api.get_fund_nav_with_date(end_date = self.end_date, fund_list = fund_list)
        self.fund_nav: pd.DataFrame = self.fund_nav.pivot_table(index='datetime', columns='fund_id', values='adjusted_net_value')
        self.fund_nav = self.fund_nav.reindex(self.trading_days.index).fillna(method='ffill').dropna(axis=0,how='all')
        self.fund_list = self.fund_nav.columns.tolist()

        self.fund_ret = np.log(self.fund_nav).diff().dropna(axis=0,how='all')
        self.fund_ret[self.fund_ret > 0.15] = 0

        # 存款利率
        bank_rate_df = self.basic_api.get_index_price(index_list=['ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d'])
        bank_rate_df = bank_rate_df.pivot_table(index='datetime', columns='index_id', values='close')
        bank_rate_df = bank_rate_df.reindex(self.trading_days.index.union(bank_rate_df.index)).fillna(method='ffill').reindex(self.trading_days.index)
        
        # 基金基准数据
        self.index_price = self.basic_api.get_index_price_dt(end_date=end_date, columns=['close'])
        self.index_price = self.index_price.pivot_table(index='datetime', columns='index_id', values='close')
        self.index_price = self.index_price.reindex(self.trading_days.index).fillna(method='ffill')
        drop_list = [i for i in self.index_price if i in bank_rate_df]
        self.index_price = self.index_price.drop(drop_list, axis=1)
        self.index_price = pd.concat([self.index_price, bank_rate_df], axis=1)
        
        # 宏观数据
        self.macroeco = RawDataApi().get_em_macroeconomic_daily(codes=['EX_DR_RATIO']).drop(columns='_update_time')
        self.macroeco = self.macroeco.pivot_table(index='datetime', columns='codes', values='value')
        self.macroeco = self.macroeco.reindex(self.trading_days.index.union(self.macroeco.index)).fillna(method='ffill').reindex(self.trading_days.index)

        # 大类指数价格
        self.fund_type_benchmark = self.index_price[[self.REPLACE_DICT[fund_type] for fund_type in self.FUND_CLASSIFIER]]
        self.fund_type_benchmark = self.fund_type_benchmark.rename(columns={v : f'{k}_bench' for k, v in self.REPLACE_DICT.items()})
        self.fund_type_benchmark = np.log(self.fund_type_benchmark).diff().dropna(axis=0,how='all')

        # 规模
        self.fund_size: pd.DataFrame = self.basic_api.get_fund_size_range(end_date=end_date)
        self.fund_size = self.fund_size.pivot_table(index='datetime', columns='fund_id', values='size')
        self.fund_size = self.fund_size.reindex(self.trading_days.index.union(self.fund_size.index)).fillna(method='ffill').reindex(self.trading_days.index)

        # 基准计算
        self.fund_benchmark_df = self.basic_api.get_fund_benchmark().drop(columns='_update_time')
        self.benchmark_ret = self.get_benchmark_return()
        self.benchmark_ret = self.benchmark_ret.reindex(self.trading_days.index)
        benchmark_cols = self.benchmark_ret.columns.tolist()
        for fund_id in self.fund_nav.columns:
            benchmark_id = fund_id
            if not benchmark_id in benchmark_cols:    
                wind_class_2 = self.wind_class_2_dict[fund_id]
                select_type = self.wind_type_dict[wind_class_2]
                benchmark_ret = self.index_price[self.REPLACE_DICT[select_type]].pct_change(1).rename(benchmark_id)
                self.benchmark_ret = self.benchmark_ret.join(benchmark_ret)
        self.fund_ret = self.fund_ret.join(self.benchmark_ret, rsuffix='_b').join(self.fund_type_benchmark)


        # 获取年报/半年报的发布日期
        md = self.end_date.strftime('%m%d')
        sentry_date = pd.Series(self.REPORT_DATE_LIST)
        sentry = sentry_date[sentry_date <= md]
        # 为半年报/年报发布预留一个季度的缓冲时间
        if sentry.empty:
            real_date = datetime.date(self.end_date.year - 1, 6, 30)
        elif len(sentry) <= 2:
            real_date = datetime.date(self.end_date.year - 1, 12, 31)
        else:
            real_date = datetime.date(self.end_date.year, 6, 30)

        # 基金机构持仓
        self.fund_hold = self.basic_api.get_history_fund_size(start_date=real_date, end_date=real_date)
        self.fund_hold = self.fund_hold[['fund_id', 'institution_holds', 'hold_num']].set_index('fund_id')
        self.fund_hold = self.fund_hold.rename(columns={'institution_holds':'ins_hold'})

        # 货基相关风险数据
        self.risk_metric = self.basic_api.get_fund_hold_asset_by_id(start_date=real_date, end_date=real_date)
        self.risk_metric = self.risk_metric[['fund_id', 'first_repo_to_nav', 'avg_ptm']].set_index('fund_id')
        self.risk_metric = self.risk_metric.rename(columns={'first_repo_to_nav':'leverage','avg_ptm':'ptm'})

        # 月度化
        self.fund_ret_monthly = self.get_resample_ret(self.fund_ret)
        self.fund_ret_monthly.index = self.fund_ret_monthly.index.date

        # 周度化
        self.fund_ret_weekly = self.get_resample_ret(df=self.fund_ret, rule='1W', min_count=3).dropna(axis=0, how='all')
        self.fund_ret_weekly.index = self.fund_ret_weekly.index.date
        
        # 周期
        self.period = ['1y', '3y', '5y', 'history']
        # monthly ret
        _df = self.fund_nav.tail(20)
        self.monthly_ret = _df.iloc[-1] / _df.iloc[0]
        self.fund_info = self.fund_info.set_index('fund_id')

    def get_benchmark_return(self) -> pd.DataFrame:
        benchmark_list: Dict[str, float] = {}
        # 遍历每一只基金的benchmark进行处理
        for row in self.fund_benchmark_df.itertuples(index=False):
            values: List[pd.Series] = []
            cons: float = 0
            # 空的benchmark表明我们没有对应的指数或无法解析公式
            if row.benchmark_s:
                benchmark: Dict[str, float] = json.loads(row.benchmark_s)
                benchmark_raw: Dict[str, float] = eval(row.benchmark)
                for (index, weight), index_raw in zip(benchmark.items(), benchmark_raw.keys()):
                    if index == '1':
                        # 表示benchmark中该项为常数
                        cons += weight
                    elif index in ('ddir', 'nonor', 'tmd_1y', 'tmd_2y', 'tmd_3m', 'tmd_3y', 'tmd_5y', 'tmd_6m', 'tmd_7d'):
                        if weight == -1:
                            # 表示我们无法解析公式
                            # print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            # self.log.append((row.fund_id, index))
                            break
                        else:
                            try:
                                if index == 'nonor':
                                    # 在这里我们用超额存款准备金率替代同业存款利率
                                    ra: pd.Series = self.macroeco.loc[:, 'EX_DR_RATIO']
                                else:
                                    ra: pd.Series = self.index_price.loc[:, index]
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                # print(f'[benchmark_return] Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                # self.log.append((row.fund_id, index))
                                break
                            else:
                                values.append(ra.iloc[1:] * 0.01 * weight / self.TRADING_DAYS_PER_YEAR)
                    else:
                        if weight == -1:
                            # 表示我们无法解析公式
                            # print(f'[benchmark_return] Error: Need fix {row.fund_id} {index} {index_raw}')
                            # self.log.append((row.fund_id, index))
                            break
                        else:
                            try:
                                ra: pd.Series = self.index_price.loc[:, index]
                            except KeyError:
                                # 表示我们没有该指数的价格数据
                                # print(f'Error: Data Missing: {row.fund_id} {index} {index_raw}')
                                # self.log.append((row.fund_id, index))
                                break
                            else:
                                ra = np.log(ra).diff().iloc[1:]
                                values.append(ra * weight)
                else:
                    if values or cons:
                        the_sum: float = sum(values)
                        if cons:
                            the_sum += np.log(math.pow(1 + cons, 1 / self.TRADING_DAYS_PER_YEAR))
                        benchmark_list[row.fund_id] = the_sum

        return pd.DataFrame.from_dict(benchmark_list)

    def get_resample_ret(self, df, rule='1M', min_count=15):

        return df.set_axis(pd.to_datetime(df.index), inplace=False).resample(rule).sum(min_count=min_count)

    def get_begin_date(self, period: str):
        if period == '1y':
            return self.trading_days.tail(self.TRADING_DAYS_PER_YEAR * 1).index[0]

        if period == '3y':
             return self.trading_days.tail(self.TRADING_DAYS_PER_YEAR * 3).index[0]

        if period == '5y':
             return self.trading_days.tail(self.TRADING_DAYS_PER_YEAR * 5).index[0]

        if period == 'history':
            return  self.trading_days.index[0]

    def calc_stutzer(self, ex_return: np.ndarray):
        information_statistic = lambda theta: np.log(np.mean(np.exp(theta[0] * ex_return)))
        theta0 = [-1.]
        bounds = Bounds((-np.inf), (0))
        result = minimize(information_statistic, theta0, method='SLSQP', bounds=bounds, tol=1e-16)
        if result.success:
            information_statistic = -result.fun
            if information_statistic <= 0:
                return 0
            stutzer_index = np.abs(np.mean(ex_return)) / np.mean(ex_return) * np.sqrt(2 * information_statistic)
            return stutzer_index
        else:
            return np.nan
    
    def calc_cl_alpha_beta(self, total: np.ndarray):
        if total.shape[0] <= 1:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        X = np.array([total[:, 1], total[:, 1]]).T
        X[:, 0][X[:, 0] < 0] = 0
        X[:, 1][X[:, 1] > 0] = 0
        if np.count_nonzero(X[:, 1]) == 0:
            return {'beta':np.nan,
                    'alpha':np.nan,
                    }
        est2 = sm.OLS(total[:, 0], sm.add_constant(X, prepend=False)).fit()
        return {'beta':est2.params[0] - est2.params[1],
                'alpha':est2.params[-1],
                }

    def calc_continue_regress_v(self, df, period='monthly'):
        if period == 'monthly':
            # 半年周期
            window = 6
            yearly_multiplier = 2
            period_num = 12
        df.loc[:,'annual_ret'] = df.fund_ret.rolling(window=window).apply(lambda x : x.sum() * yearly_multiplier - self.RISK_FEE_RATE)
        df.loc[:,'annual_vol'] = df.fund_ret.rolling(window=window).apply(lambda x : x.std(ddof=1) * np.sqrt(period_num))
        df.loc[:,'sharpe'] = df.loc[:,'annual_ret'] / df.loc[:,'annual_vol']
        df = df.replace(np.Inf, 0).replace(-np.Inf, 0).dropna() # 空值sharpe赋值0
        if df.shape[0] < 6:
            return np.nan
        mod = AutoReg(endog=df['sharpe'].dropna().values, lags=1)
        res = mod.fit()
        continue_regress_v = res.params[0]
        return continue_regress_v

    def calcualtor(self, df, period='monthly'):
        if period == 'monthly':
            yearly_multiplier = 12
            period_days = self.TRADING_DAYS_PER_MONTH
        x = df.benchmark_ret.values
        y = df.fund_ret.values
        z = df.benchmark_real.values
        ex_ret_mmf = df.fund_ret - df.mmf_bench
        x = sm.add_constant(x)
        model = regression.linear_model.OLS(y,x).fit()
        x = x[:,1]
        alpha = model.params[0] * yearly_multiplier
        beta = model.params[1]
        track_err = (y - z).std(ddof=1) * np.sqrt(yearly_multiplier) 
        downside_std = y[y < (self.RISK_FEE_RATE_PER_DAY * period_days)].std(ddof=1)
        info_ratio = alpha / track_err
        treynor = (np.nanmean(y) - self.RISK_FEE_RATE_PER_DAY * period_days) / beta if beta != 0 else np.nan
        mdd_part =  df.fund_nav[:] / df.fund_nav[:].rolling(window=df.shape[0], min_periods=1).max()
        mdd = 1 - mdd_part.min()
        if np.isnan(mdd):
            mdd_len = np.nan
        else:
            mdd_date2 = mdd_part.idxmin()
            mdd_date1 = df.fund_nav[:mdd_date2].idxmax()
            mdd_len = (mdd_date2 - mdd_date1).days

        annual_ret = df.fund_ret.sum() * (yearly_multiplier / df.shape[0]) - self.RISK_FEE_RATE
        annual_vol = df.fund_ret.std(ddof=1) * np.sqrt(yearly_multiplier)
        sharpe = annual_ret / annual_vol
        calma_ratio = annual_ret / mdd if mdd != 0 else np.nan
        var = np.quantile(df.fund_ret, 0.05)
        cvar = df.fund_ret[df.fund_ret <= var].mean()
        winning_rate_mmf = len(ex_ret_mmf[ex_ret_mmf > 0]) / len(ex_ret_mmf)
        stutzer = self.calc_stutzer(df.fund_ret - self.RISK_FEE_RATE_PER_DAY * period_days) * np.sqrt(yearly_multiplier)
        cl_res = self.calc_cl_alpha_beta(df[['fund_ret','benchmark_ret']].to_numpy())
        stock_cl_alpha = cl_res['alpha']
        stock_cl_beta = cl_res['beta']
        continue_regress_v = self.calc_continue_regress_v(df, period)
        
        return {
            'alpha':alpha,
            'beta':beta,
            'track_err':track_err,
            'downside_std':downside_std,
            'info_ratio':info_ratio,
            'treynor':treynor,
            'mdd':mdd,
            'mdd_len':mdd_len,
            'annual_vol':annual_vol,
            'annual_ret':annual_ret,
            'sharpe':sharpe,
            'calma_ratio':calma_ratio,
            'var':var,
            'cvar':cvar,
            'winning_rate_mmf':winning_rate_mmf,
            'stutzer':stutzer,
            'stock_cl_alpha':stock_cl_alpha,
            'stock_cl_beta':stock_cl_beta,
            'continue_regress_v':continue_regress_v,
        }

    def loop_item(self, fund_id):
        _t0 = time.time()
        idx = self.fund_list.index(fund_id)
        wind_class_2 = self.wind_class_2_dict[fund_id]
        type_class = self.wind_type_dict[wind_class_2]
        cols = [fund_id,f'{fund_id}_b',f'{type_class}_bench']
        fund_data = self.fund_ret_monthly[cols]
        fund_data['fund_nav'] = np.exp(fund_data[fund_id].cumsum())
        fund_data = fund_data.rename(columns={fund_id:'fund_ret',f'{type_class}_bench':'benchmark_ret',f'{fund_id}_b':'benchmark_real'})
        fund_data = fund_data.join(self.fund_ret_monthly['mmf_bench'])
        fund_result = {
            'fund_id': fund_id,
            'datetime': self.end_date
        }
        for period in self.period:
            begin_date = self.get_begin_date(period)
            fund_data_period =  fund_data.loc[begin_date:].copy().dropna()
            if all(fund_data_period.fund_ret == 0):
                continue
            period_result = self.calcualtor(fund_data_period, 'monthly')
            period_result = { f'{k}~{period}_m' : v for k, v in period_result.items() }
            fund_result.update(period_result)

        cols = [fund_id, f'{type_class}_bench']
        fund_data = self.fund_ret_weekly[cols]
        df = fund_data.rename(columns={fund_id:'fund_ret',f'{type_class}_bench':'benchmark_ret'}).dropna()
        cl_res = self.calc_cl_alpha_beta(df[['fund_ret','benchmark_ret']].to_numpy())
        stock_cl_alpha = cl_res['alpha']
        stock_cl_beta = cl_res['beta']
        fund_result.update({'stock_cl_alpha~history_w':stock_cl_alpha})
        fund_result.update({'stock_cl_beta~history_w':stock_cl_beta})
        _t1 = time.time()
        #print(f'{fund_id} {idx} / {len(self.fund_list)} cost time {_t1 - _t0}')
        return fund_result

    def calc_indicator(self):
        _t00 = time.time()
        self.result = []
        self.fund_size_columns = self.fund_size.columns.tolist()
        self.fund_hold_columns = self.fund_hold.columns.tolist()
        self.risk_funds_list = self.risk_metric.index
        for fund_id in self.fund_list:
            try:
                self.result.append(self.loop_item(fund_id))
            except:
                print(f'boom fund_id {fund_id}')
        self.result_df = pd.DataFrame(self.result)
        self.result_df = pd.merge(self.result_df, self.fund_hold, on=['fund_id'])
        self.result_df = pd.merge(self.result_df, self.risk_metric, on=['fund_id']).set_index('fund_id')
        self.result_df.loc[:,'prn_hold'] = 100 - self.result_df['ins_hold']
        self.result_df.loc[:,'scale'] = self.fund_size.loc[self.end_date]
        self.result_df.loc[:,'monthly_ret'] = self.monthly_ret
        self.result_df.loc[:,'total_ret'] = np.exp(self.fund_ret.sum())
        self.result_df.loc[:,'mdd~history_daily'] = 1 - (self.fund_nav / self.fund_nav.cummax()).min()
        self.result_df = self.result_df.reset_index()

        self.fund_info.loc[:,'trade_days'] = [i.days for i in (self.end_date - self.fund_info.start_date)]
        self.fund_info.loc[:,'fund_type'] = self.fund_info.wind_class_2.map(self.wind_type_dict)
        self.result_df = self.result_df.set_index('fund_id')
        self.result_df = self.result_df.join(self.fund_info[['wind_class_2','fund_type','desc_name']])
        self.result_df = self.result_df.replace(np.Inf,np.nan).replace(-np.Inf,np.nan)
        self.result_df= self.result_df.join(self.fund_info[['trade_days']])
        self.indicator = self.result_df.copy()
        _t11 = time.time()
        print(f'total cost time {_t11 - _t00}')

    def calc_score(self):
        res = []
        for fund_type in self.FUND_CLASSIFIER:
            df = self.indicator[self.indicator['fund_type'] == fund_type].copy()
            ret = 0.1 * normalize(df['annual_ret~1y_m']) + 0.1 * normalize(df['annual_ret~3y_m']) + 0.3 * normalize(df['annual_ret~5y_m']) + 0.2 * normalize(df['annual_ret~history_m'])
            if fund_type == 'stock':
                #df.loc[:,'ret_ability'] = score_rescale(0.6 * normalize(df['alpha~history_m']) + 0.4 * ret)
                df.loc[:,'ret_ability'] = score_rescale( 0.4 * score_rescale(df['annual_ret~history_m']) + 0.6 * score_rescale(df['total_ret']))
                df.loc[:,'risk_ability'] = score_rescale(0.3 * score_rescale(-df['mdd~history_daily']) + 0.4 * score_rescale(-df['annual_vol~history_m']) + 0.3 * score_rescale(-df['downside_std~history_m']))
                df.loc[:,'stable_ability'] = score_rescale(0.4 * score_rescale(df['trade_days']) + 0.4 * score_rescale(df['continue_regress_v~history_m']) + 0.2 * score_rescale(df['scale']))
                df.loc[:,'select_time'] =  score_rescale(normalize(df['stock_cl_beta~history_w']))
                df.loc[:,'select_stock'] = score_rescale(normalize(df['stock_cl_alpha~history_w']))
                df.loc[:,'total_score'] = score_rescale(df.loc[:,'ret_ability'] + df.loc[:,'risk_ability'] + df.loc[:,'stable_ability'] + df.loc[:,'select_time'] + df.loc[:,'select_stock'] )
            
            elif fund_type == 'bond':
                df.loc[:,'ret_ability'] = score_rescale( 0.8 * score_rescale(df['annual_ret~history_m']) + 0.2 * score_rescale(df['total_ret']))
                df.loc[:,'risk_ability'] = score_rescale(0.3 * score_rescale(-df['mdd~history_daily']) + 0.4 * score_rescale(-df['annual_vol~history_m']) + 0.3 * score_rescale(-df['downside_std~history_m']))
                df.loc[:,'stable_ability'] = score_rescale(0.4 * score_rescale(df['trade_days']) + 0.4 * score_rescale(df['continue_regress_v~history_m']) + 0.2 * score_rescale(df['scale']))
                df.loc[:,'select_time'] =  score_rescale(normalize(df['stock_cl_beta~history_w']))
                df.loc[:,'select_stock'] = score_rescale(normalize(df['stock_cl_alpha~history_w']))
                df.loc[:,'total_score'] = score_rescale(df.loc[:,'ret_ability'] + df.loc[:,'risk_ability'] + df.loc[:,'stable_ability'] + df.loc[:,'select_time'] + df.loc[:,'select_stock'] )
            
            elif fund_type == 'index':
                df.loc[:,'ret_ability'] = score_rescale( 0.4 * score_rescale(df['annual_ret~history_m']) + 0.6 * score_rescale(df['total_ret']))
                #df.loc[:,'track_ability'] = score_rescale(-normalize(df['track_err~history_m']))
                df.loc[:,'risk_ability'] = score_rescale(0.3 * score_rescale(-df['mdd~history_daily']) + 0.4 * score_rescale(-df['annual_vol~history_m']) + 0.3 * score_rescale(-df['downside_std~history_m']))
                df.loc[:,'stable_ability'] = score_rescale(0.4 * score_rescale(df['trade_days']) + 0.4 * score_rescale(df['continue_regress_v~history_m']) + 0.2 * score_rescale(df['scale']))
                df.loc[:,'total_score'] = score_rescale(df.loc[:,'ret_ability'] + df.loc[:,'risk_ability'] + df.loc[:,'stable_ability'] )
            
            elif fund_type == 'QDII':
                df.loc[:,'ret_ability'] = score_rescale( 0.4 * score_rescale(df['annual_ret~history_m']) + 0.6 * score_rescale(df['total_ret']))
                df.loc[:,'risk_ability'] = score_rescale(0.3 * score_rescale(-df['mdd~history_daily']) + 0.4 * score_rescale(-df['annual_vol~history_m']) + 0.3 * score_rescale(-df['downside_std~history_m']))
                df.loc[:,'stable_ability'] = score_rescale(-normalize(df['annual_vol~history_m']))
                df.loc[:,'select_time'] = np.nan
                df.loc[:,'select_stock'] = np.nan
                df.loc[:,'total_score'] = score_rescale(df.loc[:,'ret_ability'] + df.loc[:,'risk_ability'] + df.loc[:,'stable_ability'] )
            
            elif fund_type == 'mmf':
                df.loc[:,'ret_ability'] = score_rescale(normalize(df['monthly_ret']))
                df.loc[:,'risk_ability'] = score_rescale(normalize(df['prn_hold']) + normalize(df['scale']))
                df.loc[:,'total_score'] = score_rescale(df.loc[:,'ret_ability'] + df.loc[:,'risk_ability'] )
            res.append(df)
        self.fund_indicator_score = pd.concat(res)

    def get_mng_best_fund(self):
        desc_dic = self.fund_info[['fund_id','desc_name']].set_index('fund_id').to_dict()['desc_name']
        manager_info = self.fund_manager_info[self.fund_manager_info.end  == datetime.date(2040,1,1)].copy()
        mng_name_dic = self.fund_manager_info.reset_index()[['manager_id','name']].set_index('manager_id').to_dict()['name']
        res = []
        for fund_type in self.mng_best_fund:
            _df = self.mng_best_fund['stock'].tail(1).reset_index().drop(columns=['datetime']).dropna(axis=1).replace(desc_dic).rename(columns=mng_name_dic).T
            _df = _df.reset_index().rename(columns={'index':'mng_name',0:'desc_name'})
            _df.loc[:,'fund_type'] = fund_type
            res.append(_df)
        self.mng_best_fund_now = pd.concat(res)

    def score_rescale(self, df):
        return df.rank(pct=True) * 100

    def add_mng_score(self):
        bucket_name = 'tl-fund-dm'
        _s3_bucket_uri: str = f's3://{bucket_name}/dm'
        name = 'mng_indicator_score'
        s3_uri = f'{_s3_bucket_uri}/{name}.parquet'
        score = pd.read_parquet(s3_uri)
        
        mng_info = DerivedDataApi().get_fund_manager_info().set_index('mng_id')
        mng_dic = mng_info[mng_info.end_date == datetime.date(2040,1,1)]['fund_id']

        # 找基金对应基金经理分数
        fund_type_list = list(self.fund_indicator_score.fund_type.unique())
        res = []
        for fund_type in fund_type_list:
            _score = score[score.fund_type == fund_type].copy().dropna(subset=['total_score'])
            _fund_indicator_part = self.fund_indicator_score[self.fund_indicator_score.fund_type == fund_type]#[['total_score']]
            _score = _score.set_index('mng_id').join(mng_dic)
            _score = _score.reset_index().set_index('fund_id').rename(columns={'total_score':'mng_score'})[['mng_score']]
            _score = _score.reset_index().groupby('fund_id').max()
            _fund_indicator_part = _fund_indicator_part.drop(columns='mng_score',errors='ignore').join(_score['mng_score'])
            res.append(_fund_indicator_part)
        fund_indicator = pd.concat(res,axis=0)
        dic = {
            'stock':['ret_ability','risk_ability','stable_ability','select_time','select_stock','mng_score'],
            'bond':['ret_ability','risk_ability','stable_ability','select_time','select_stock','mng_score'],
            'index':['ret_ability','risk_ability','stable_ability','mng_score'],
            'QDII':['ret_ability','risk_ability','stable_ability','mng_score'],
            'mmf':['ret_ability','risk_ability','mng_score']
        }
        res = []
        for fund_type, score_list in dic.items():
            df = fund_indicator[fund_indicator['fund_type'] == fund_type].copy()
            df['total_score'] = score_rescale(df[score_list].sum(axis=1))
            res.append(df)
        fund_indicator_result = pd.concat(res)
        self.fund_indicator_score = fund_indicator_result

    def to_s3(self):
        bucket_name = 'tl-fund-dm'
        _s3_bucket_uri: str = f's3://{bucket_name}/dm'
        name = 'fund_indicator_score'
        self.fund_indicator_score.to_parquet(f'{_s3_bucket_uri}/{name}.parquet', compression='gzip')
        
    def process(self,end_date):
        failed_tasks = []
        try:
            self.init(end_date=end_date)
            self.calc_indicator()
            self.calc_score()
            self.add_mng_score()
            self.to_s3()
        except Exception as e:
            print(e)
            traceback.print_exc()
            failed_tasks.append('manager_indicator_score')
        return failed_tasks


from cassandra.cqlengine import columns
from cassandra.cqlengine.models import Model
from ...wrapper.cas.cas_connection import CassandraConnector
from ...wrapper.cas.constants import CAS_KEY_SPACE

CassandraConnector().get_conn()


class TagMethodIndexLevel(Model):
    __table_name__ = 'tag_method_index_level'
    __keyspace__ = CAS_KEY_SPACE
    __connection__ = 'surfing'

    index_id = columns.Text(primary_key=True)                       # 指数ID
    data = columns.List(columns.Map(columns.Text, columns.Text))    # 数据
    update_time = columns.DateTime()                                # 更新时间

    def to_dict(self):
        return {
            'index_id': self.index_id,
            'data': self.data,
            'update_time': self.update_time,
        }





from ....manager.data_tables import StockInfoDataTables
from .reader import (
    read_stock_dividend,
    read_stock_price_sheet,
    read_stock_equity_sheet,
    read_financial_real_publish_date,
    read_balance_sheet,
    read_income_sheet,
    read_income_quarterly_sheet,
    read_cash_flow_sheet,
    read_cash_flow_quarterly_sheet,
)
from .financial_ttm import handle_income_ttm, handle_cash_flow_ttm
from .handle_financial_ratios import compute_ttm_quarterly_ratios
from .handle_prices_ratios import compute_value_ratios
from .compute_ratings import compute_price_ratings, compute_finance_ratings


def init_stock_data(stock_id: str) -> StockInfoDataTables:
    sdt = StockInfoDataTables()
    sdt.balance = read_balance_sheet(stock_id, [])

    sdt.income = read_income_sheet(stock_id, [])
    sdt.income_q = read_income_quarterly_sheet(stock_id, [])
    sdt.income_ttm = handle_income_ttm(sdt.income_q)

    sdt.cash_flow_q = read_cash_flow_quarterly_sheet(stock_id, [])
    sdt.cash_flow_ttm = handle_cash_flow_ttm(sdt.cash_flow_q)

    sdt.dividend = read_stock_dividend(stock_id)
    sdt.prices = read_stock_price_sheet(stock_id)
    sdt.equity = read_stock_equity_sheet(stock_id)
    sdt.real_dates = read_financial_real_publish_date(stock_id)

    return sdt


def clean_up_sdt(sdt: StockInfoDataTables):
    sdt.real_dates = None
    sdt.prices = None
    sdt.dividend = None
    sdt.equity = None


def generate_stock_factor_dict(sdt: StockInfoDataTables):
    sdt.factor_dict = {}

    if len(sdt.values) > 0:
        df = sdt.values
        df = df.sort_index()
        sdt.factor_dict.update(df.iloc[-1].to_dict())
        sdt.factor_dict['评估日期'] = df.index[-1]
    if len(sdt.finance) > 0:
        df = sdt.finance
        df = df.sort_index()
        sdt.factor_dict.update(df.iloc[-1].to_dict())


def compute_stock(param):
    stock_id = param.get('stock_id')
    assert stock_id, 'No stock id'
    try:
        sdt = init_stock_data(stock_id)
        sdt.finance = compute_ttm_quarterly_ratios(sdt)
        sdt.values = compute_value_ratios(sdt)
        sdt.values = compute_price_ratings(sdt.values)
        sdt.finance = compute_finance_ratings(sdt.finance)

        clean_up_sdt(sdt)
        generate_stock_factor_dict(sdt)
        return {stock_id: sdt}
    except Exception:
        import traceback
        print(stock_id)
        print(traceback.format_exc())
        return {}








